"""
Attributes are arbitrary data stored on objects. Attributes supports
both pure-string values and pickled arbitrary data.

Attributes are also used to implement Nicks. This module also contains
the Attribute- and NickHandlers as well as the `NAttributeHandler`,
which is a non-db version of Attributes.


"""
import fnmatch
import re
from collections import defaultdict

from django.conf import settings
from django.db import models
from django.utils.encoding import smart_str

from evennia.locks.lockhandler import LockHandler
from evennia.utils.dbserialize import from_pickle, to_pickle
from evennia.utils.idmapper.models import SharedMemoryModel
from evennia.utils.picklefield import PickledObjectField
from evennia.utils.utils import is_iter, lazy_property, make_iter, to_str

_TYPECLASS_AGGRESSIVE_CACHE = settings.TYPECLASS_AGGRESSIVE_CACHE

# -------------------------------------------------------------
#
#   Attributes
#
# -------------------------------------------------------------


class IAttribute:
    """
    Attributes are things that are specific to different types of objects. For
    example, a drink container needs to store its fill level, whereas an exit
    needs to store its open/closed/locked/unlocked state. These are done via
    attributes, rather than making different classes for each object type and
    storing them directly. The added benefit is that we can add/remove
    attributes on the fly as we like.

    The Attribute class defines the following properties:
     - key (str): Primary identifier.
     - lock_storage (str): Perm strings.
     - model (str): A string defining the model this is connected to. This
        is a natural_key, like "objects.objectdb"
     - date_created (datetime): When the attribute was created.
     - value (any): The data stored in the attribute, in pickled form
        using wrappers to be able to store/retrieve models.
     - strvalue (str): String-only data. This data is not pickled and
        is thus faster to search for in the database.
     - category (str): Optional character string for grouping the
        Attribute.

    This class is an API/Interface/Abstract base class; do not instantiate it directly.
    """

    @lazy_property
    def locks(self):
        return LockHandler(self)

    key = property(lambda self: self.db_key)
    strvalue = property(lambda self: self.db_strvalue)
    category = property(lambda self: self.db_category)
    model = property(lambda self: self.db_model)
    attrtype = property(lambda self: self.db_attrtype)
    date_created = property(lambda self: self.db_date_created)

    def __lock_storage_get(self):
        return self.db_lock_storage

    def __lock_storage_set(self, value):
        self.db_lock_storage = value

    def __lock_storage_del(self):
        self.db_lock_storage = ""

    lock_storage = property(__lock_storage_get, __lock_storage_set, __lock_storage_del)

    def access(self, accessing_obj, access_type="read", default=False, **kwargs):
        """
        Determines if another object has permission to access.

        Args:
            accessing_obj (object): Entity trying to access this one.
            access_type (str, optional): Type of access sought, see
                the lock documentation.
            default (bool, optional): What result to return if no lock
                of access_type was found. The default, `False`, means a lockdown
                policy, only allowing explicit access.
            kwargs (any, optional): Not used; here to make the API consistent with
                other access calls.

        Returns:
            result (bool): If the lock was passed or not.

        """
        result = self.locks.check(accessing_obj, access_type=access_type, default=default)
        return result

    #
    #
    # Attribute methods
    #
    #

    def __str__(self):
        return smart_str("%s(%s)" % (self.db_key, self.id))

    def __repr__(self):
        return "%s(%s)" % (self.db_key, self.id)


class InMemoryAttribute(IAttribute):
    """
    This Attribute is used purely for NAttributes/NAttributeHandler. It has no database backend.

    """

    # Primary Key has no meaning for an InMemoryAttribute. This merely serves to satisfy other code.

    def __init__(self, pk, **kwargs):
        """
        Create an Attribute that exists only in Memory.

        Args:
            pk (int): This is a fake 'primary key' / id-field. It doesn't actually have to be
                unique, but is fed an incrementing number from the InMemoryBackend by default. This
                is needed only so Attributes can be sorted. Some parts of the API also see the lack
                of a .pk field as a sign that the Attribute was deleted.
            **kwargs: Other keyword arguments are used to construct the actual Attribute.

        """
        self.id = pk
        self.pk = pk

        # Copy all kwargs to local properties. We use db_ for compatability here.
        for key, value in kwargs.items():
            # Value and locks are special. We must call the wrappers.
            if key == "value":
                self.value = value
            elif key == "lock_storage":
                self.lock_storage = value
            else:
                setattr(self, f"db_{key}", value)

    # value property (wraps db_value)
    def __value_get(self):
        return self.db_value

    def __value_set(self, new_value):
        self.db_value = new_value

    def __value_del(self):
        pass

    value = property(__value_get, __value_set, __value_del)


class AttributeProperty:
    """
    AttributeProperty.

    """

    attrhandler_name = "attributes"

    def __init__(self, default=None, category=None, strattr=False, lockstring="", autocreate=True):
        """
        Allows for specifying Attributes as Django-like 'fields' on the class level. Note that while
        one can set a lock on the Attribute, there is no way to *check* said lock when accessing via
        the property - use the full `AttributeHandler` if you need to do access checks. Note however
        that if you use the full `AttributeHandler` to access this Attribute, the `at_get/at_set`
        methods on this class will _not_ fire (because you are bypassing the `AttributeProperty`
        entirely in that case).

        Initialize an Attribute as a property descriptor.

        Keyword Args:
            default (any): A default value if the attr is not set. If a callable, this will be
                run without any arguments and is expected to return the default value.
            category (str): The attribute's category. If unset, use class default.
            strattr (bool): If set, this Attribute *must* be a simple string, and will be
                stored more efficiently.
            lockstring (str): This is not itself useful with the property, but only if
                using the full AttributeHandler.get(accessing_obj=...) to access the
                Attribute.
            autocreate (bool): True by default; this means Evennia makes sure to create a new
                copy of the Attribute (with the default value) whenever a new object with this
                property is created. If `False`, no Attribute will be created until the property
                is explicitly assigned a value. This makes it more efficient while it retains
                its default (there's no db access), but without an actual Attribute generated,
                one cannot access it via .db, the AttributeHandler or see it with `examine`.
        Example:
        ::

            class Character(DefaultCharacter):
                foo = AttributeProperty(default="Bar")

        """
        self._default = default
        self._category = category
        self._strattr = strattr
        self._lockstring = lockstring
        self._autocreate = autocreate
        self._key = ""

    @property
    def _default(self):
        """
        Tries returning a new instance of default if callable.

        """
        if callable(self.__default):
            return self.__default()

        return self.__default

    @_default.setter
    def _default(self, value):
        self.__default = value

    def __set_name__(self, cls, name):
        """
        Called when descriptor is first assigned to the class. It is called with
        the name of the field.

        """
        self._key = name

    def __get__(self, instance, owner):
        """
        Called when the attrkey is retrieved from the instance.

        """
        value = self._default
        try:
            value = self.at_get(
                getattr(instance, self.attrhandler_name).get(
                    key=self._key,
                    default=self._default,
                    category=self._category,
                    strattr=self._strattr,
                    raise_exception=self._autocreate,
                ),
                instance,
            )
        except AttributeError:
            if self._autocreate:
                # attribute didn't exist and autocreate is set
                self.__set__(instance, self._default)
            else:
                raise
        return value

    def __set__(self, instance, value):
        """
        Called when assigning to the property (and when auto-creating an Attribute).

        """
        (
            getattr(instance, self.attrhandler_name).add(
                self._key,
                self.at_set(value, instance),
                category=self._category,
                lockstring=self._lockstring,
                strattr=self._strattr,
            )
        )

    def __delete__(self, instance):
        """
        Called when running `del` on the property. Will remove/clear the Attribute. Note that
        the Attribute will be recreated next retrieval unless the AttributeProperty is also
        removed in code!

        """
        getattr(instance, self.attrhandler_name).remove(key=self._key, category=self._category)

    def at_set(self, value, obj):
        """
        The value to set is passed through the method. It can be used to customize/validate
        the input in a custom child class.

        Args:
            value (any): The value about to the stored in this Attribute.
            obj (object): Object the attribute is attached to

        Returns:
            any: The value to store.

        Raises:
            AttributeError: If the value is invalid to store.

        Notes:
            This is will only fire if you actually set the Attribute via this `AttributeProperty`.
            That is, if you instead set it via the `AttributeHandler` (or via `.db`), you are
            bypassing this `AttributeProperty` entirely and this method is never reached.

        """
        return value

    def at_get(self, value, obj):
        """
        The value returned from the Attribute is passed through this method. It can be used
        to react to the retrieval or modify the result in some way.

        Args:
            value (any): Value returned from the Attribute.
            obj (object): Object the attribute is attached to

        Returns:
            any: The value to return to the caller.

        Notes:
            This is will only fire if you actually get the Attribute via this `AttributeProperty`.
            That is, if you instead get it via the `AttributeHandler` (or via `.db`), you are
            bypassing this `AttributeProperty` entirely and this method is never reached.

        """
        return value


class NAttributeProperty(AttributeProperty):
    """
    NAttribute property descriptor. Allows for specifying NAttributes as Django-like 'fields'
    on the class level.

    Example:
    ::

        class Character(DefaultCharacter):
            foo = NAttributeProperty(default="Bar")

    """

    attrhandler_name = "nattributes"


class Attribute(IAttribute, SharedMemoryModel):
    """
    This attribute is stored via Django. Most Attributes will be using this class.

    """

    #
    # Attribute Database Model setup
    #
    # These database fields are all set using their corresponding properties,
    # named same as the field, but without the db_* prefix.
    db_key = models.CharField("key", max_length=255, db_index=True)
    db_value = PickledObjectField(
        "value",
        null=True,
        help_text=(
            "The data returned when the attribute is accessed. Must be "
            "written as a Python literal if editing through the admin "
            "interface. Attribute values which are not Python literals "
            "cannot be edited through the admin interface."
        ),
    )
    db_strvalue = models.TextField(
        "strvalue", null=True, blank=True, help_text="String-specific storage for quick look-up"
    )
    db_category = models.CharField(
        "category",
        max_length=128,
        db_index=True,
        blank=True,
        null=True,
        help_text="Optional categorization of attribute.",
    )
    # Lock storage
    db_lock_storage = models.TextField(
        "locks", blank=True, help_text="Lockstrings for this object are stored here."
    )
    db_model = models.CharField(
        "model",
        max_length=32,
        db_index=True,
        blank=True,
        null=True,
        help_text=(
            "Which model of object this attribute is attached to (A "
            "natural key like 'objects.objectdb'). You should not change "
            "this value unless you know what you are doing."
        ),
    )
    # subclass of Attribute (None or nick)
    db_attrtype = models.CharField(
        "attrtype",
        max_length=16,
        db_index=True,
        blank=True,
        null=True,
        help_text="Subclass of Attribute (None or nick)",
    )
    # time stamp
    db_date_created = models.DateTimeField("date_created", editable=False, auto_now_add=True)

    # Database manager
    # objects = managers.AttributeManager()

    class Meta:
        "Define Django meta options"
        verbose_name = "Attribute"

    # Wrapper properties to easily set database fields. These are
    # @property decorators that allows to access these fields using
    # normal python operations (without having to remember to save()
    # etc). So e.g. a property 'attr' has a get/set/del decorator
    # defined that allows the user to do self.attr = value,
    # value = self.attr and del self.attr respectively (where self
    # is the object in question).

    # lock_storage wrapper. Overloaded for saving to database.
    def __lock_storage_get(self):
        return self.db_lock_storage

    def __lock_storage_set(self, value):
        self.db_lock_storage = value
        self.save(update_fields=["db_lock_storage"])

    def __lock_storage_del(self):
        self.db_lock_storage = ""
        self.save(update_fields=["db_lock_storage"])

    lock_storage = property(__lock_storage_get, __lock_storage_set, __lock_storage_del)

    # value property (wraps db_value)
    @property
    def value(self):
        """
        Getter. Allows for `value = self.value`.
        We cannot cache here since it makes certain cases (such
        as storing a dbobj which is then deleted elsewhere) out-of-sync.
        The overhead of unpickling seems hard to avoid.
        """
        return from_pickle(self.db_value, db_obj=self)

    @value.setter
    def value(self, new_value):
        """
        Setter. Allows for self.value = value. We cannot cache here,
        see self.__value_get.
        """
        self.db_value = to_pickle(new_value)
        self.save(update_fields=["db_value"])

    @value.deleter
    def value(self):
        """Deleter. Allows for del attr.value. This removes the entire attribute."""
        self.delete()


#
# Handlers making use of the Attribute model
#


class IAttributeBackend:
    """
    Abstract interface for the backends used by the Attribute Handler.

    All Backends must implement this base class.
    """

    _attrcreate = "attrcreate"
    _attredit = "attredit"
    _attrread = "attrread"
    _attrclass = None

    def __init__(self, handler, attrtype):
        self.handler = handler
        self.obj = handler.obj
        self._attrtype = attrtype
        self._objid = handler.obj.id
        self._cache = {}
        # store category names fully cached
        self._catcache = {}
        # full cache was run on all attributes
        self._cache_complete = False

    def query_all(self):
        """
        Fetch all Attributes from this object.

        Returns:
            attrlist (list): A list of Attribute objects.
        """
        raise NotImplementedError()

    def query_key(self, key, category):
        """

        Args:
            key (str): The key of the Attribute being searched for.
            category (str or None): The category of the desired Attribute.

        Returns:
            attribute (IAttribute): A single Attribute.
        """
        raise NotImplementedError()

    def query_category(self, category):
        """
        Returns every matching Attribute as a list, given a category.

        This method calls up whatever storage the backend uses.

        Args:
            category (str or None): The category to query.

        Returns:
            attrs (list): The discovered Attributes.
        """
        raise NotImplementedError()

    def _full_cache(self):
        """Cache all attributes of this object"""
        if not _TYPECLASS_AGGRESSIVE_CACHE:
            return
        attrs = self.query_all()
        self._cache = {
            f"{to_str(attr.key).lower()}-{attr.category.lower() if attr.category else None}": attr
            for attr in attrs
        }
        self._cache_complete = True

    def _get_cache_key(self, key, category):
        """
        Fetch cache key.

        Args:
            key (str): The key of the Attribute being searched for.
            category (str or None): The category of the desired Attribute.

        Returns:
            attribute (IAttribute): A single Attribute.
        """
        cachekey = "%s-%s" % (key, category)
        cachefound = False
        try:
            attr = _TYPECLASS_AGGRESSIVE_CACHE and self._cache[cachekey]
            cachefound = True
        except KeyError:
            attr = None

        if attr and (not hasattr(attr, "pk") and attr.pk is None):
            # clear out Attributes deleted from elsewhere. We must search this anew.
            attr = None
            cachefound = False
            del self._cache[cachekey]
        if cachefound and _TYPECLASS_AGGRESSIVE_CACHE:
            if attr:
                return [attr]  # return cached entity
            else:
                return []  # no such attribute: return an empty list
        else:
            conn = self.query_key(key, category)
            if conn:
                attr = conn[0].attribute
                if _TYPECLASS_AGGRESSIVE_CACHE:
                    self._cache[cachekey] = attr
                return [attr] if attr.pk else []
            else:
                # There is no such attribute. We will explicitly save that
                # in our cache to avoid firing another query if we try to
                # retrieve that (non-existent) attribute again.
                if _TYPECLASS_AGGRESSIVE_CACHE:
                    self._cache[cachekey] = None
                return []

    def _get_cache_category(self, category):
        """
        Retrieves Attribute list (by category) from cache.

        Args:
            category (str or None): The category to query.

        Returns:
            attrs (list): The discovered Attributes.
        """
        catkey = "-%s" % category
        if _TYPECLASS_AGGRESSIVE_CACHE and catkey in self._catcache:
            return [attr for key, attr in self._cache.items() if key.endswith(catkey) and attr]
        else:
            # we have to query to make this category up-date in the cache
            attrs = self.query_category(category)
            if _TYPECLASS_AGGRESSIVE_CACHE:
                for attr in attrs:
                    if attr.pk:
                        cachekey = "%s-%s" % (attr.key, category)
                        self._cache[cachekey] = attr
                # mark category cache as up-to-date
                self._catcache[catkey] = True
            return attrs

    def _get_cache(self, key=None, category=None):
        """
        Retrieve from cache or database (always caches)

        Args:
            key (str, optional): Attribute key to query for
            category (str, optional): Attribiute category

        Returns:
            args (list): Returns a list of zero or more matches
                found from cache or database.
        Notes:
            When given a category only, a search for all objects
            of that cateogory is done and the category *name* is
            stored. This tells the system on subsequent calls that the
            list of cached attributes of this category is up-to-date
            and that the cache can be queried for category matches
            without missing any.
            The TYPECLASS_AGGRESSIVE_CACHE=False setting will turn off
            caching, causing each attribute access to trigger a
            database lookup.

        """
        key = key.strip().lower() if key else None
        category = category.strip().lower() if category is not None else None
        if key:
            return self._get_cache_key(key, category)
        return self._get_cache_category(category)

    def get(self, key=None, category=None):
        """
        Frontend for .get_cache. Retrieves Attribute(s).

        Args:
            key (str, optional): Attribute key to query for
            category (str, optional): Attribiute category

        Returns:
            args (list): Returns a list of zero or more matches
                found from cache or database.
        """
        return self._get_cache(key, category)

    def _set_cache(self, key, category, attr_obj):
        """
        Update cache.

        Args:
            key (str): A cleaned key string
            category (str or None): A cleaned category name
            attr_obj (IAttribute): The newly saved attribute

        """
        if not _TYPECLASS_AGGRESSIVE_CACHE:
            return
        if not key:  # don't allow an empty key in cache
            return
        cachekey = "%s-%s" % (key, category)
        catkey = "-%s" % category
        self._cache[cachekey] = attr_obj
        # mark that the category cache is no longer up-to-date
        self._catcache.pop(catkey, None)
        self._cache_complete = False

    def _delete_cache(self, key, category):
        """
        Remove attribute from cache

        Args:
            key (str): A cleaned key string
            category (str or None): A cleaned category name

        """
        catkey = "-%s" % category
        if key:
            cachekey = "%s-%s" % (key, category)
            self._cache.pop(cachekey, None)
        else:
            self._cache = {
                key: attrobj
                for key, attrobj in list(self._cache.items())
                if not key.endswith(catkey)
            }
        # mark that the category cache is no longer up-to-date
        self._catcache.pop(catkey, None)
        self._cache_complete = False

    def reset_cache(self):
        """
        Reset cache from the outside.
        """
        self._cache_complete = False
        self._cache = {}
        self._catcache = {}

    def do_create_attribute(self, key, category, lockstring, value, strvalue):
        """
        Does the hard work of actually creating Attributes, whatever is needed.

        Args:
            key (str): The Attribute's key.
            category (str or None): The Attribute's category, or None
            lockstring (str): Any locks for the Attribute.
            value (obj): The Value of the Attribute.
            strvalue (bool): Signifies if this is a strvalue Attribute. Value MUST be a string or
                this will lead to Trouble. Ignored for InMemory attributes.

        Returns:
            attr (IAttribute): The new Attribute.
        """
        raise NotImplementedError()

    def create_attribute(self, key, category, lockstring, value, strvalue=False, cache=True):
        """
        Creates Attribute (using the class specified for the backend), (optionally) caches it, and
        returns it.

        This MUST actively save the Attribute to whatever database backend is used, AND
        call self.set_cache(key, category, new_attrobj)

        Args:
            key (str): The Attribute's key.
            category (str or None): The Attribute's category, or None
            lockstring (str): Any locks for the Attribute.
            value (obj): The Value of the Attribute.
            strvalue (bool): Signifies if this is a strvalue Attribute. Value MUST be a string or
                this will lead to Trouble. Ignored for InMemory attributes.
            cache (bool): Whether to cache the new Attribute

        Returns:
            attr (IAttribute): The new Attribute.
        """
        attr = self.do_create_attribute(key, category, lockstring, value, strvalue)
        if cache:
            self._set_cache(key, category, attr)
        return attr

    def do_update_attribute(self, attr, value, strvalue):
        """
        Simply sets a new Value to an Attribute.

        Args:
            attr (IAttribute): The Attribute being changed.
            value (obj): The Value for the Attribute.
            strvalue (bool): If True, `value` is expected to be a string.

        """
        raise NotImplementedError()

    def do_batch_update_attribute(self, attr_obj, category, lock_storage, new_value, strvalue):
        """
        Called opnly by batch add. For the database backend, this is a method
        of updating that can alter category and lock-storage.

        Args:
            attr_obj (IAttribute): The Attribute being altered.
            category (str or None): The attribute's (new) category.
            lock_storage (str): The attribute's new locks.
            new_value (obj): The Attribute's new value.
            strvalue (bool): Signifies if this is a strvalue Attribute. Value MUST be a string or
                this will lead to Trouble. Ignored for InMemory attributes.
        """
        raise NotImplementedError()

    def do_batch_finish(self, attr_objs):
        """
        Called after batch_add completed. Used for handling database operations
        and/or caching complications.

        Args:
            attr_objs (list of IAttribute): The Attributes created/updated thus far.

        """
        raise NotImplementedError()

    def batch_add(self, *args, **kwargs):
        """
        Batch-version of `.add()`. This is more efficient than repeat-calling
        `.add` when having many Attributes to add.

        Args:
             *args (tuple): Tuples of varying length representing the
                Attribute to add to this object. Supported tuples are

                - (key, value)
                - (key, value, category)
                - (key, value, category, lockstring)
                - (key, value, category, lockstring, default_access)

        Raises:
            RuntimeError: If trying to pass a non-iterable as argument.

        Notes:
            The indata tuple order matters, so if you want a lockstring but no
            category, set the category to `None`. This method does not have the
            ability to check editing permissions and is mainly used internally.
            It does not use the normal `self.add` but applies the Attributes
            directly to the database.

        """
        new_attrobjs = []
        strattr = kwargs.get("strattr", False)
        for tup in args:
            if not is_iter(tup) or len(tup) < 2:
                raise RuntimeError("batch_add requires iterables as arguments (got %r)." % tup)
            ntup = len(tup)
            keystr = str(tup[0]).strip().lower()
            new_value = tup[1]
            category = str(tup[2]).strip().lower() if ntup > 2 and tup[2] is not None else None
            lockstring = tup[3] if ntup > 3 else ""

            attr_objs = self._get_cache(keystr, category)

            if attr_objs:
                attr_obj = attr_objs[0]
                # update an existing attribute object
                self.do_batch_update_attribute(attr_obj, category, lockstring, new_value, strattr)
            else:
                new_attr = self.do_create_attribute(
                    keystr, category, lockstring, new_value, strvalue=strattr
                )
                new_attrobjs.append(new_attr)
        if new_attrobjs:
            self.do_batch_finish(new_attrobjs)

    def do_delete_attribute(self, attr):
        """
        Does the hard work of actually deleting things.

        Args:
            attr (IAttribute): The attribute to delete.
        """
        raise NotImplementedError()

    def delete_attribute(self, attr):
        """
        Given an Attribute, deletes it. Also remove it from cache.

        Args:
            attr (IAttribute): The attribute to delete.
        """
        if not attr:
            return
        self._delete_cache(attr.key, attr.category)
        self.do_delete_attribute(attr)

    def update_attribute(self, attr, value, strattr=False):
        """
        Simply updates an Attribute.

        Args:
            attr (IAttribute): The attribute to delete.
            value (obj): The new value.
            strattr (bool): If set, the `value` is a raw string.
        """
        self.do_update_attribute(attr, value, strattr)

    def do_batch_delete(self, attribute_list):
        """
        Given a list of attributes, deletes them all.
        The default implementation is fine, but this is overridable since some databases may allow
        for a better method.

        Args:
            attribute_list (list of IAttribute):
        """
        for attribute in attribute_list:
            self.delete_attribute(attribute)

    def clear_attributes(self, category, accessing_obj, default_access):
        """
        Remove all Attributes on this object.

        Args:
            category (str, optional): If given, clear only Attributes
                of this category.
            accessing_obj (object, optional): If given, check the
                `attredit` lock on each Attribute before continuing.
            default_access (bool, optional): Use this permission as
                fallback if `access_obj` is given but there is no lock of
                type `attredit` on the Attribute in question.

        """
        category = category.strip().lower() if category is not None else None

        if not self._cache_complete:
            self._full_cache()

        if category is not None:
            attrs = [attr for attr in self._cache.values() if attr.category == category]
        else:
            attrs = self._cache.values()

        if accessing_obj:
            self.do_batch_delete(
                [
                    attr
                    for attr in attrs
                    if attr.access(accessing_obj, self._attredit, default=default_access)
                ]
            )
        else:
            # have to cast the results to a list or we'll get a RuntimeError for removing from the
            # dict we're iterating
            self.do_batch_delete(list(attrs))
        self.reset_cache()

    def get_all_attributes(self):
        """
        Simply returns all Attributes of this object, sorted by their IDs.

        Returns:
            attributes (list of IAttribute)
        """
        if _TYPECLASS_AGGRESSIVE_CACHE:
            if not self._cache_complete:
                self._full_cache()
            return sorted([attr for attr in self._cache.values() if attr], key=lambda o: o.id)
        else:
            return sorted([attr for attr in self.query_all() if attr], key=lambda o: o.id)


class InMemoryAttributeBackend(IAttributeBackend):
    """
    This Backend for Attributes stores NOTHING in the database. Everything is kept in memory, and
    normally lost on a crash, reload, shared memory flush, etc. It generates IDs for the Attributes
    it manages, but these are of little importance beyond sorting and satisfying the caching logic
    to know an Attribute hasn't been deleted out from under the cache's nose.

    """

    _attrclass = InMemoryAttribute

    def __init__(self, handler, attrtype):
        super().__init__(handler, attrtype)
        self._storage = dict()
        self._category_storage = defaultdict(list)
        self._id_counter = 0

    def _next_id(self):
        """
        Increments the internal ID counter and returns the new value.

        Returns:
            next_id (int): A simple integer.
        """
        self._id_counter += 1
        return self._id_counter

    def query_all(self):
        return self._storage.values()

    def query_key(self, key, category):
        found = self._storage.get((key, category), None)
        if found:
            return [found]
        return []

    def query_category(self, category):
        if category is None:
            return self._storage.values()
        return self._category_storage.get(category, [])

    def do_create_attribute(self, key, category, lockstring, value, strvalue):
        """
        See parent class.

        strvalue has no meaning for InMemory attributes.

        """
        new_attr = self._attrclass(
            pk=self._next_id(), key=key, category=category, lock_storage=lockstring, value=value
        )
        self._storage[(key, category)] = new_attr
        self._category_storage[category].append(new_attr)
        return new_attr

    def do_update_attribute(self, attr, value, strvalue):
        attr.value = value

    def do_batch_update_attribute(self, attr_obj, category, lock_storage, new_value, strvalue):
        """
        No need to bother saving anything. Just set some values.
        """
        attr_obj.db_category = category
        attr_obj.db_lock_storage = lock_storage if lock_storage else ""
        attr_obj.value = new_value

    def do_batch_finish(self, attr_objs):
        """
        Nothing to do here for In-Memory.

        Args:
            attr_objs (list of IAttribute): The Attributes created/updated thus far.
        """
        pass

    def do_delete_attribute(self, attr):
        """
        Removes the Attribute from local storage. Once it's out of the cache, garbage collection
        will handle the rest.

        Args:
            attr (IAttribute): The attribute to delete.
        """
        del self._storage[(attr.key, attr.category)]
        self._category_storage[attr.category].remove(attr)


class ModelAttributeBackend(IAttributeBackend):
    """
    Uses Django models for storing Attributes.
    """

    _attrclass = Attribute
    _m2m_fieldname = "db_attributes"

    def __init__(self, handler, attrtype):
        super().__init__(handler, attrtype)
        self._model = to_str(handler.obj.__dbclass__.__name__.lower())

    def query_all(self):
        query = {
            "%s__id" % self._model: self._objid,
            "attribute__db_model__iexact": self._model,
            "attribute__db_attrtype": self._attrtype,
        }
        return [
            conn.attribute
            for conn in getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query)
        ]

    def query_key(self, key, category):
        query = {
            "%s__id" % self._model: self._objid,
            "attribute__db_model__iexact": self._model,
            "attribute__db_attrtype": self._attrtype,
            "attribute__db_key__iexact": key.lower(),
            "attribute__db_category__iexact": category.lower() if category else None,
        }
        if not self.obj.pk:
            return []
        return getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query)

    def query_category(self, category):
        query = {
            "%s__id" % self._model: self._objid,
            "attribute__db_model__iexact": self._model,
            "attribute__db_attrtype": self._attrtype,
            "attribute__db_category__iexact": category.lower() if category else None,
        }
        return [
            conn.attribute
            for conn in getattr(self.obj, self._m2m_fieldname).through.objects.filter(**query)
        ]

    def do_create_attribute(self, key, category, lockstring, value, strvalue):
        kwargs = {
            "db_key": key,
            "db_category": category,
            "db_model": self._model,
            "db_lock_storage": lockstring if lockstring else "",
            "db_attrtype": self._attrtype,
        }
        if strvalue:
            kwargs["db_value"] = None
            kwargs["db_strvalue"] = value
        else:
            kwargs["db_value"] = to_pickle(value)
            kwargs["db_strvalue"] = None
        new_attr = self._attrclass(**kwargs)
        new_attr.save()
        getattr(self.obj, self._m2m_fieldname).add(new_attr)
        self._set_cache(key, category, new_attr)
        return new_attr

    def do_update_attribute(self, attr, value, strvalue):
        if strvalue:
            attr.value = None
            attr.db_strvalue = value
        else:
            attr.value = value
            attr.db_strvalue = None
        attr.save(update_fields=["db_strvalue", "db_value"])

    def do_batch_update_attribute(self, attr_obj, category, lock_storage, new_value, strvalue):
        attr_obj.db_category = category
        attr_obj.db_lock_storage = lock_storage if lock_storage else ""
        if strvalue:
            # store as a simple string (will not notify OOB handlers)
            attr_obj.db_strvalue = new_value
            attr_obj.value = None
        else:
            # store normally (this will also notify OOB handlers)
            attr_obj.value = new_value
            attr_obj.db_strvalue = None
        attr_obj.save(update_fields=["db_strvalue", "db_value", "db_category", "db_lock_storage"])

    def do_batch_finish(self, attr_objs):
        # Add new objects to m2m field all at once
        getattr(self.obj, self._m2m_fieldname).add(*attr_objs)

    def do_delete_attribute(self, attr):
        try:
            attr.delete()
        except AssertionError:
            # This could happen if the Attribute has already been deleted.
            pass


class AttributeHandler:
    """
    Handler for adding Attributes to the object.
    """

    _attrcreate = "attrcreate"
    _attredit = "attredit"
    _attrread = "attrread"
    _attrtype = None

    def __init__(self, obj, backend_class):
        """
        Setup the AttributeHandler.

        Args:
            obj (TypedObject): An Account, Object, Channel, ServerSession (not technically a typed
                object), etc.  backend_class (IAttributeBackend class): The class of the backend to
                use.
        """
        self.obj = obj
        self.backend = backend_class(self, self._attrtype)

    def has(self, key=None, category=None):
        """
        Checks if the given Attribute (or list of Attributes) exists on
        the object.

        Args:
            key (str or iterable): The Attribute key or keys to check for.
                If `None`, search by category.
            category (str or None): Limit the check to Attributes with this
                category (note, that `None` is the default category).

        Returns:
            has_attribute (bool or list): If the Attribute exists on
                this object or not. If `key` was given as an iterable then
                the return is a list of booleans.

        """
        ret = []
        category = category.strip().lower() if category is not None else None
        for keystr in make_iter(key):
            keystr = key.strip().lower()
            ret.extend(bool(attr) for attr in self.backend.get(keystr, category))
        return ret[0] if len(ret) == 1 else ret

    def get(
        self,
        key=None,
        default=None,
        category=None,
        return_obj=False,
        strattr=False,
        raise_exception=False,
        accessing_obj=None,
        default_access=True,
        return_list=False,
    ):
        """
        Get the Attribute.

        Args:
            key (str or list, optional): the attribute identifier or
                multiple attributes to get. if a list of keys, the
                method will return a list.
            default (any, optional): The value to return if an
                Attribute was not defined. If set, it will be returned in
                a one-item list.
            category (str, optional): the category within which to
                retrieve attribute(s).
            return_obj (bool, optional): If set, the return is not the value of the
                Attribute but the Attribute object itself.
            strattr (bool, optional): Return the `strvalue` field of
                the Attribute rather than the usual `value`, this is a
                string-only value for quick database searches.
            raise_exception (bool, optional): When an Attribute is not
                found, the return from this is usually `default`. If this
                is set, an exception is raised instead.
            accessing_obj (object, optional): If set, an `attrread`
                permission lock will be checked before returning each
                looked-after Attribute.
            default_access (bool, optional): If no `attrread` lock is set on
                object, this determines if the lock should then be passed or not.
            return_list (bool, optional): Always return a list, also if there is only
                one or zero matches found.

        Returns:
            result (any or list): One or more matches for keys and/or
                categories. Each match will be the value of the found Attribute(s)
                unless `return_obj` is True, at which point it will be the
                attribute object itself or None. If `return_list` is True, this
                will always be a list, regardless of the number of elements.

        Raises:
            AttributeError: If `raise_exception` is set and no matching Attribute
                was found matching `key`.

        """

        ret = []
        for keystr in make_iter(key):
            # it's okay to send a None key
            attr_objs = self.backend.get(keystr, category)
            if attr_objs:
                ret.extend(attr_objs)
            elif raise_exception:
                raise AttributeError
            elif return_obj:
                ret.append(None)

        if accessing_obj:
            # check 'attrread' locks
            ret = [
                attr
                for attr in ret
                if attr.access(accessing_obj, self._attrread, default=default_access)
            ]
        if strattr:
            ret = ret if return_obj else [attr.strvalue for attr in ret if attr]
        else:
            ret = ret if return_obj else [attr.value for attr in ret if attr]

        if return_list:
            return ret if ret else [default] if default is not None else []
        return ret[0] if ret and len(ret) == 1 else ret or default

    def add(
        self,
        key,
        value,
        category=None,
        lockstring="",
        strattr=False,
        accessing_obj=None,
        default_access=True,
    ):
        """
        Add attribute to object, with optional `lockstring`.

        Args:
            key (str): An Attribute name to add.
            value (any or str): The value of the Attribute. If
                `strattr` keyword is set, this *must* be a string.
            category (str, optional): The category for the Attribute.
                The default `None` is the normal category used.
            lockstring (str, optional): A lock string limiting access
                to the attribute.
            strattr (bool, optional): Make this a string-only Attribute.
                This is only ever useful for optimization purposes.
            accessing_obj (object, optional): An entity to check for
                the `attrcreate` access-type. If not passing, this method
                will be exited.
            default_access (bool, optional): What access to grant if
                `accessing_obj` is given but no lock of the type
                `attrcreate` is defined on the Attribute in question.

        """
        if accessing_obj and not self.obj.access(
            accessing_obj, self._attrcreate, default=default_access
        ):
            # check create access
            return

        if not key:
            return

        category = category.strip().lower() if category is not None else None
        keystr = key.strip().lower()
        attr_obj = self.backend.get(key, category)

        if attr_obj:
            # update an existing attribute object
            attr_obj = attr_obj[0]
            self.backend.update_attribute(attr_obj, value, strattr)
        else:
            # create a new Attribute (no OOB handlers can be notified)
            self.backend.create_attribute(keystr, category, lockstring, value, strattr)

    def batch_add(self, *args, **kwargs):
        """
        Batch-version of `add()`. This is more efficient than
        repeat-calling add when having many Attributes to add.

        Args:
            *args (tuple): Each argument should be a tuples (can be of varying
                length) representing the Attribute to add to this object.
                Supported tuples are

                - (key, value)
                - (key, value, category)
                - (key, value, category, lockstring)
                - (key, value, category, lockstring, default_access)

        Keyword Args:
            strattr (bool): If `True`, value must be a string. This
                will save the value without pickling which is less
                flexible but faster to search (not often used except
                internally).

        Raises:
            RuntimeError: If trying to pass a non-iterable as argument.

        Notes:
            The indata tuple order matters, so if you want a lockstring
            but no category, set the category to `None`. This method
            does not have the ability to check editing permissions like
            normal .add does, and is mainly used internally. It does not
            use the normal self.add but apply the Attributes directly
            to the database.

        """
        self.backend.batch_add(*args, **kwargs)

    def remove(
        self,
        key=None,
        category=None,
        raise_exception=False,
        accessing_obj=None,
        default_access=True,
    ):
        """
        Remove attribute or a list of attributes from object.

        Args:
            key (str or list, optional): An Attribute key to remove or a list of keys. If
                multiple keys, they must all be of the same `category`. If None and
                category is not given, remove all Attributes.
            category (str, optional): The category within which to
                remove the Attribute.
            raise_exception (bool, optional): If set, not finding the
                Attribute to delete will raise an exception instead of
                just quietly failing.
            accessing_obj (object, optional): An object to check
                against the `attredit` lock. If not given, the check will
                be skipped.
            default_access (bool, optional): The fallback access to
                grant if `accessing_obj` is given but there is no
                `attredit` lock set on the Attribute in question.

        Raises:
            AttributeError: If `raise_exception` is set and no matching Attribute
                was found matching `key`.

        Notes:
            If neither key nor category is given, this acts as clear().

        """

        if key is None:
            self.clear(
                category=category, accessing_obj=accessing_obj, default_access=default_access
            )
            return

        category = category.strip().lower() if category is not None else None

        for keystr in make_iter(key):
            keystr = keystr.lower()

            attr_objs = self.backend.get(keystr, category)
            for attr_obj in attr_objs:
                if not (
                    accessing_obj
                    and not attr_obj.access(accessing_obj, self._attredit, default=default_access)
                ):
                    self.backend.delete_attribute(attr_obj)
            if not attr_objs and raise_exception:
                raise AttributeError

    def clear(self, category=None, accessing_obj=None, default_access=True):
        """
        Remove all Attributes on this object.

        Args:
            category (str, optional): If given, clear only Attributes
                of this category.
            accessing_obj (object, optional): If given, check the
                `attredit` lock on each Attribute before continuing.
            default_access (bool, optional): Use this permission as
                fallback if `access_obj` is given but there is no lock of
                type `attredit` on the Attribute in question.

        """
        self.backend.clear_attributes(category, accessing_obj, default_access)

    def all(self, category=None, accessing_obj=None, default_access=True):
        """
        Return all Attribute objects on this object, regardless of category.

        Args:
            category (str, optional): A given category to limit results to.
            accessing_obj (object, optional): Check the `attrread`
                lock on each attribute before returning them. If not
                given, this check is skipped.
            default_access (bool, optional): Use this permission as a
                fallback if `accessing_obj` is given but one or more
                Attributes has no lock of type `attrread` defined on them.

        Returns:
            Attributes (list): All the Attribute objects (note: Not
                their values!) in the handler.

        """
        attrs = self.backend.get_all_attributes()
        if category:
            attrs = [attr for attr in attrs if attr.category == category]

        if accessing_obj:
            return [
                attr
                for attr in attrs
                if attr.access(accessing_obj, self._attrread, default=default_access)
            ]
        else:
            return attrs

    def reset_cache(self):
        self.backend.reset_cache()


# DbHolders for .db and .ndb properties on Typeclasses.

_GA = object.__getattribute__
_SA = object.__setattr__


class DbHolder:
    "Holder for allowing property access of attributes"

    def __init__(self, obj, name, manager_name="attributes"):
        _SA(self, name, _GA(obj, manager_name))
        _SA(self, "name", name)

    def __getattribute__(self, attrname):
        if attrname == "all":
            # we allow to overload our default .all
            attr = _GA(self, _GA(self, "name")).get("all")
            return attr if attr else _GA(self, "all")
        return _GA(self, _GA(self, "name")).get(attrname)

    def __setattr__(self, attrname, value):
        _GA(self, _GA(self, "name")).add(attrname, value)

    def __delattr__(self, attrname):
        _GA(self, _GA(self, "name")).remove(attrname)

    def get_all(self):
        return _GA(self, _GA(self, "name")).backend.get_all_attributes()

    all = property(get_all)


#
# Nick templating
#

"""
This supports the use of replacement templates in nicks:

This happens in two steps:

1) The user supplies a template that is converted to a regex according
   to the unix-like templating language.
2) This regex is tested against nicks depending on which nick replacement
   strategy is considered (most commonly inputline).
3) If there is a template match and there are templating markers,
   these are replaced with the arguments actually given.

@desc $1 $2 $3

This will be converted to the following regex:

    \@desc (?P<1>\w+) (?P<2>\w+) $(?P<3>\w+)

Supported template markers (through fnmatch)
   *       matches anything (non-greedy)     -> .*?
   ?       matches any single character      ->
   [seq]   matches any entry in sequence
   [!seq]  matches entries not in sequence
Custom arg markers
   $N      argument position (1-99)

"""
_RE_OR = re.compile(r"(?<!\\)\|")
_RE_NICK_RE_ARG = re.compile(r"arg([1-9][0-9]?)")
_RE_NICK_ARG = re.compile(r"\\(\$)([1-9][0-9]?)")
_RE_NICK_RAW_ARG = re.compile(r"(\$)([1-9][0-9]?)")
_RE_NICK_SPACE = re.compile(r"\\ ")


class NickTemplateInvalid(ValueError):
    pass


def initialize_nick_templates(pattern, replacement, pattern_is_regex=False):
    """
    Initialize the nick templates for matching and remapping a string.

    Args:
        pattern (str): The pattern to be used for nick recognition. This will
            be parsed for shell patterns into a regex, unless `pattern_is_regex`
            is `True`, in which case it must be an already valid regex string. In
            this case, instead of `$N`, numbered arguments must instead be given
            as matching groups named as `argN`, such as `(?P<arg1>.+?)`.
        replacement (str): The template to be used to replace the string
            matched by the pattern. This can contain `$N` markers and is never
            parsed into a regex.
        pattern_is_regex (bool): If set, `pattern` is a full regex string
            instead of containing shell patterns.

    Returns:
        regex, template  (str): Regex to match against strings and template
            with markers ``{arg1}, {arg2}``, etc for replacement using the standard
            `.format` method.

    Raises:
        evennia.typecalasses.attributes.NickTemplateInvalid: If the in/out
        template does not have a matching number of `$args`.

    Examples:
        - `pattern` (shell syntax): `"grin $1"`
        - `pattern` (regex): `"grin (?P<arg1.+?>)"`
        - `replacement`: `"emote gives a wicked grin to $1"`

    """

    # create the regex from the pattern
    if pattern_is_regex:
        # Note that for a regex we can't validate in the way we do for the shell
        # pattern, since you may have complex OR statements or optional arguments.

        # Explicit regex given from the onset - this already contains argN
        # groups.  we need to split out any | - separated parts so we can
        # attach the line-break/ending extras all regexes require.
        pattern_regex_string = r"|".join(
            or_part + r"(?:[\n\r]*?)\Z" for or_part in _RE_OR.split(pattern)
        )

    else:
        # Shell pattern syntax - convert $N to argN groups
        # for the shell pattern we make sure we have matching $N on both sides
        pattern_args = [match.group(1) for match in _RE_NICK_RAW_ARG.finditer(pattern)]
        replacement_args = [match.group(1) for match in _RE_NICK_RAW_ARG.finditer(replacement)]
        if set(pattern_args) != set(replacement_args):
            # We don't have the same amount of argN/$N tags in input/output.
            raise NickTemplateInvalid("Nicks: Both in/out-templates must contain the same $N tags.")

        # generate regex from shell pattern
        pattern_regex_string = fnmatch.translate(pattern)
        pattern_regex_string = _RE_NICK_SPACE.sub(r"\\s+", pattern_regex_string)
        pattern_regex_string = _RE_NICK_ARG.sub(
            lambda m: "(?P<arg%s>.+?)" % m.group(2), pattern_regex_string
        )
        # we must account for a possible line break coming over the wire
        pattern_regex_string = pattern_regex_string[:-2] + r"(?:[\n\r]*?)\Z"

    # map the replacement to match the arg1 group-names, to make replacement easy
    replacement_string = _RE_NICK_RAW_ARG.sub(lambda m: "{arg%s}" % m.group(2), replacement)

    return pattern_regex_string, replacement_string


def parse_nick_template(string, template_regex, outtemplate):
    """
    Parse a text using a template and map it to another template

    Args:
        string (str): The input string to process
        template_regex (regex): A template regex created with
            initialize_nick_template.
        outtemplate (str): The template to which to map the matches
            produced by the template_regex. This should have $1, $2,
            etc to match the template-regex. Un-found $N-markers (possible if
            the regex has optional matching groups) are replaced with empty
            strings.

    """
    match = template_regex.match(string)
    if match:
        matchdict = {
            key: value if value is not None else "" for key, value in match.groupdict().items()
        }
        return True, outtemplate.format_map(matchdict)
    return False, string


class NickHandler(AttributeHandler):
    """
    Handles the addition and removal of Nicks. Nicks are special
    versions of Attributes with an `_attrtype` hardcoded to `nick`.
    They also always use the `strvalue` fields for their data.

    """

    _attrtype = "nick"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._regex_cache = {}

    def has(self, key, category="inputline"):
        """
        Args:
            key (str or iterable): The Nick key or keys to check for.
            category (str): Limit the check to Nicks with this
                category (note, that `None` is the default category).

        Returns:
            has_nick (bool or list): If the Nick exists on this object
                or not. If `key` was given as an iterable then the return
                is a list of booleans.

        """
        return super().has(key, category=category)

    def get(self, key=None, category="inputline", return_tuple=False, **kwargs):
        """
        Get the replacement value matching the given key and category

        Args:
            key (str or list, optional): the attribute identifier or
                multiple attributes to get. if a list of keys, the
                method will return a list.
            category (str, optional): the category within which to
                retrieve the nick. The "inputline" means replacing data
                sent by the user.
            return_tuple (bool, optional): return the full nick tuple rather
                than just the replacement. For non-template nicks this is just
                a string.
            kwargs (any, optional): These are passed on to `AttributeHandler.get`.

        Returns:
            str or tuple:  The nick replacement string or nick tuple.

        """
        if return_tuple or "return_obj" in kwargs:
            return super().get(key=key, category=category, **kwargs)
        else:
            retval = super().get(key=key, category=category, **kwargs)
            if retval:
                return (
                    retval[3]
                    if isinstance(retval, tuple)
                    else [tup[3] for tup in make_iter(retval)]
                )
            return None

    def add(self, pattern, replacement, category="inputline", pattern_is_regex=False, **kwargs):
        """
        Add a new nick, a mapping pattern -> replacement.

        Args:
            pattern (str): A pattern to match for. This will be parsed for
                shell patterns using the `fnmatch` library and can contain
                `$N`-markers to indicate the locations of arguments to catch. If
                `pattern_is_regex=True`, this must instead be a valid regular
                expression and the `$N`-markers must be named `argN` that matches
                numbered regex groups (see examples).
            replacement (str): The string (or template) to replace `key` with
                (the "nickname"). This may contain `$N` markers to indicate where to
                place the argument-matches
            category (str, optional): the category within which to
                retrieve the nick. The "inputline" means replacing data
                sent by the user.
            pattern_is_regex (bool): If `True`, the `pattern` will be parsed as a
                raw regex string. Instead of using `$N` markers in this string, one
                then must mark numbered arguments as a named regex-groupd named `argN`.
                For example, `(?P<arg1>.+?)` will match the behavior of using `$1`
                in the shell pattern.
            **kwargs (any, optional): These are passed on to `AttributeHandler.get`.

        Notes:
            For most cases, the shell-pattern is much shorter and easier. The
            regex pattern form can be useful for more complex matchings though,
            for example in order to add optional arguments, such as with
            `(?P<argN>.*?)`.

        Example:
            - pattern (default shell syntax): `"gr $1 at $2"`
            - pattern (with pattern_is_regex=True): `r"gr (?P<arg1>.+?) at (?P<arg2>.+?)"`
            - replacement: `"emote With a flourish, $1 grins at $2."`

        """
        nick_regex, nick_template = initialize_nick_templates(
            pattern, replacement, pattern_is_regex=pattern_is_regex
        )
        super().add(
            pattern, (nick_regex, nick_template, pattern, replacement), category=category, **kwargs
        )

    def remove(self, key, category="inputline", **kwargs):
        """
        Remove Nick with matching category.

        Args:
            key (str): A key for the nick to match for.
            category (str, optional): the category within which to
                removethe nick. The "inputline" means replacing data
                sent by the user.
            kwargs (any, optional): These are passed on to `AttributeHandler.get`.

        """
        super().remove(key, category=category, **kwargs)

    def nickreplace(self, raw_string, categories=("inputline", "channel"), include_account=True):
        """
        Apply nick replacement of entries in raw_string with nick replacement.

        Args:
            raw_string (str): The string in which to perform nick
                replacement.
            categories (tuple, optional): Replacement categories in
                which to perform the replacement, such as "inputline",
                "channel" etc.
            include_account (bool, optional): Also include replacement
                with nicks stored on the Account level.
            kwargs (any, optional): Not used.

        Returns:
            string (str): A string with matching keys replaced with
                their nick equivalents.

        """
        nicks = {}
        for category in make_iter(categories):
            nicks.update(
                {
                    nick.key: nick
                    for nick in make_iter(self.get(category=category, return_obj=True))
                    if nick and nick.key
                }
            )
        if include_account and self.obj.has_account:
            for category in make_iter(categories):
                nicks.update(
                    {
                        nick.key: nick
                        for nick in make_iter(
                            self.obj.account.nicks.get(category=category, return_obj=True)
                        )
                        if nick and nick.key
                    }
                )
        for key, nick in nicks.items():
            nick_regex, template, _, _ = nick.value
            regex = self._regex_cache.get(nick_regex)
            if not regex:
                try:
                    regex = re.compile(nick_regex, re.I + re.DOTALL + re.U)
                except re.error:
                    from evennia.utils import logger

                    logger.log_trace("Probably nick being created with unvalidated regex mapping.")
                    continue
                self._regex_cache[nick_regex] = regex

            is_match, raw_string = parse_nick_template(raw_string, regex, template)
            if is_match:
                break
        return raw_string



"""
Extended Room - Griatch 2012, vincent-lg 2019, Griatch 2023

"""

from .extended_room import CmdExtendedRoomDesc  # noqa
from .extended_room import CmdExtendedRoomDetail  # noqa
from .extended_room import CmdExtendedRoomGameTime  # noqa
from .extended_room import CmdExtendedRoomLook  # noqa
from .extended_room import CmdExtendedRoomState  # noqa
from .extended_room import ExtendedRoom  # noqa
from .extended_room import ExtendedRoomCmdSet  # noqa



"""
The command template for the default MUX-style command set. There
is also an Account/OOC version that makes sure caller is an Account object.
"""

from evennia.commands.command import Command
from evennia.utils import utils

# limit symbol import for API
__all__ = ("MuxCommand", "MuxAccountCommand")


class MuxCommand(Command):
    """
    This sets up the basis for a MUX command. The idea
    is that most other Mux-related commands should just
    inherit from this and don't have to implement much
    parsing of their own unless they do something particularly
    advanced.

    Note that the class's __doc__ string (this text) is
    used by Evennia to create the automatic help entry for
    the command, so make sure to document consistently here.
    """

    def has_perm(self, srcobj):
        """
        This is called by the cmdhandler to determine
        if srcobj is allowed to execute this command.
        We just show it here for completeness - we
        are satisfied using the default check in Command.
        """
        return super().has_perm(srcobj)

    def at_pre_cmd(self):
        """
        This hook is called before self.parse() on all commands
        """
        pass

    def at_post_cmd(self):
        """
        This hook is called after the command has finished executing
        (after self.func()).
        """
        pass

    def parse(self):
        """
        This method is called by the cmdhandler once the command name
        has been identified. It creates a new set of member variables
        that can be later accessed from self.func() (see below)

        The following variables are available for our use when entering this
        method (from the command definition, and assigned on the fly by the
        cmdhandler):
           self.key - the name of this command ('look')
           self.aliases - the aliases of this cmd ('l')
           self.permissions - permission string for this command
           self.help_category - overall category of command

           self.caller - the object calling this command
           self.cmdstring - the actual command name used to call this
                            (this allows you to know which alias was used,
                             for example)
           self.args - the raw input; everything following self.cmdstring.
           self.cmdset - the cmdset from which this command was picked. Not
                         often used (useful for commands like 'help' or to
                         list all available commands etc)
           self.obj - the object on which this command was defined. It is often
                         the same as self.caller.

        A MUX command has the following possible syntax:

          name[ with several words][/switch[/switch..]] arg1[,arg2,...] [[=|,] arg[,..]]

        The 'name[ with several words]' part is already dealt with by the
        cmdhandler at this point, and stored in self.cmdname (we don't use
        it here). The rest of the command is stored in self.args, which can
        start with the switch indicator /.

        Optional variables to aid in parsing, if set:
          self.switch_options  - (tuple of valid /switches expected by this
                                  command (without the /))
          self.rhs_split       - Alternate string delimiter or tuple of strings
                                 to separate left/right hand sides. tuple form
                                 gives priority split to first string delimiter.

        This parser breaks self.args into its constituents and stores them in the
        following variables:
          self.switches = [list of /switches (without the /)]
          self.raw = This is the raw argument input, including switches
          self.args = This is re-defined to be everything *except* the switches
          self.lhs = Everything to the left of = (lhs:'left-hand side'). If
                     no = is found, this is identical to self.args.
          self.rhs: Everything to the right of = (rhs:'right-hand side').
                    If no '=' is found, this is None.
          self.lhslist - [self.lhs split into a list by comma]
          self.rhslist - [list of self.rhs split into a list by comma]
          self.arglist = [list of space-separated args (stripped, including '=' if it exists)]

          All args and list members are stripped of excess whitespace around the
          strings, but case is preserved.
        """
        raw = self.args
        args = raw.strip()
        # Without explicitly setting these attributes, they assume default values:
        if not hasattr(self, "switch_options"):
            self.switch_options = None
        if not hasattr(self, "rhs_split"):
            self.rhs_split = "="
        if not hasattr(self, "account_caller"):
            self.account_caller = False

        # split out switches
        switches, delimiters = [], self.rhs_split
        if self.switch_options:
            self.switch_options = [opt.lower() for opt in self.switch_options]
        if args and len(args) > 1 and raw[0] == "/":
            # we have a switch, or a set of switches. These end with a space.
            switches = args[1:].split(None, 1)
            if len(switches) > 1:
                switches, args = switches
                switches = switches.split("/")
            else:
                args = ""
                switches = switches[0].split("/")
            # If user-provides switches, parse them with parser switch options.
            if switches and self.switch_options:
                valid_switches, unused_switches, extra_switches = [], [], []
                for element in switches:
                    option_check = [opt for opt in self.switch_options if opt == element]
                    if not option_check:
                        option_check = [
                            opt for opt in self.switch_options if opt.startswith(element)
                        ]
                    match_count = len(option_check)
                    if match_count > 1:
                        extra_switches.extend(
                            option_check
                        )  # Either the option provided is ambiguous,
                    elif match_count == 1:
                        valid_switches.extend(option_check)  # or it is a valid option abbreviation,
                    elif match_count == 0:
                        unused_switches.append(element)  # or an extraneous option to be ignored.
                if extra_switches:  # User provided switches
                    self.msg(
                        "|g%s|n: |wAmbiguous switch supplied: Did you mean /|C%s|w?"
                        % (self.cmdstring, " |nor /|C".join(extra_switches))
                    )
                if unused_switches:
                    plural = "" if len(unused_switches) == 1 else "es"
                    self.msg(
                        '|g%s|n: |wExtra switch%s "/|C%s|w" ignored.'
                        % (self.cmdstring, plural, "|n, /|C".join(unused_switches))
                    )
                switches = valid_switches  # Only include valid_switches in command function call
        arglist = [arg.strip() for arg in args.split()]

        # check for arg1, arg2, ... = argA, argB, ... constructs
        lhs, rhs = args.strip(), None
        if lhs:
            if delimiters and hasattr(delimiters, "__iter__"):  # If delimiter is iterable,
                best_split = delimiters[0]  # (default to first delimiter)
                for this_split in delimiters:  # try each delimiter
                    if this_split in lhs:  # to find first successful split
                        best_split = this_split  # to be the best split.
                        break
            else:
                best_split = delimiters
            # Parse to separate left into left/right sides using best_split delimiter string
            if best_split in lhs:
                lhs, rhs = lhs.split(best_split, 1)
        # Trim user-injected whitespace
        rhs = rhs.strip() if rhs is not None else None
        lhs = lhs.strip()
        # Further split left/right sides by comma delimiter
        lhslist = [arg.strip() for arg in lhs.split(",")] if lhs is not None else []
        rhslist = [arg.strip() for arg in rhs.split(",")] if rhs is not None else []
        # save to object properties:
        self.raw = raw
        self.switches = switches
        self.args = args.strip()
        self.arglist = arglist
        self.lhs = lhs
        self.lhslist = lhslist
        self.rhs = rhs
        self.rhslist = rhslist

        # if the class has the account_caller property set on itself, we make
        # sure that self.caller is always the account if possible. We also create
        # a special property "character" for the puppeted object, if any. This
        # is convenient for commands defined on the Account only.
        if self.account_caller:
            if utils.inherits_from(self.caller, "evennia.objects.objects.DefaultObject"):
                # caller is an Object/Character
                self.character = self.caller
                self.caller = self.caller.account
            elif utils.inherits_from(self.caller, "evennia.accounts.accounts.DefaultAccount"):
                # caller was already an Account
                self.character = self.caller.get_puppet(self.session)
            else:
                self.character = None

    def get_command_info(self):
        """
        Update of parent class's get_command_info() for MuxCommand.
        """
        variables = "\n".join(
            " |w{}|n ({}): {}".format(key, type(val), val) for key, val in self.__dict__.items()
        )
        string = f"""
Command {self} has no defined `func()` - showing on-command variables: No child func() defined for {self} - available variables:
{variables}
        """
        self.msg(string)
        # a simple test command to show the available properties
        string = "-" * 50
        string += f"\n|w{self.key}|n - Command variables from evennia:\n"
        string += "-" * 50
        string += f"\nname of cmd (self.key): |w{self.key}|n\n"
        string += f"cmd aliases (self.aliases): |w{self.aliases}|n\n"
        string += f"cmd locks (self.locks): |w{self.locks}|n\n"
        string += f"help category (self.help_category): |w{self.help_category}|n\n"
        string += f"object calling (self.caller): |w{self.caller}|n\n"
        string += f"object storing cmdset (self.obj): |w{self.obj}|n\n"
        string += f"command string given (self.cmdstring): |w{self.cmdstring}|n\n"
        # show cmdset.key instead of cmdset to shorten output
        string += utils.fill(f"current cmdset (self.cmdset): |w{self.cmdset}|n\n")
        string += "\n" + "-" * 50
        string += "\nVariables from MuxCommand baseclass\n"
        string += "-" * 50
        string += f"\nraw argument (self.raw): |w{self.raw}|n \n"
        string += f"cmd args (self.args): |w{self.args}|n\n"
        string += f"cmd switches (self.switches): |w{self.switches}|n\n"
        string += f"cmd options (self.switch_options): |w{self.switch_options}|n\n"
        string += f"cmd parse left/right using (self.rhs_split): |w{self.rhs_split}|n\n"
        string += f"space-separated arg list (self.arglist): |w{self.arglist}|n\n"
        string += f"lhs, left-hand side of '=' (self.lhs): |w{self.lhs}|n\n"
        string += f"lhs, comma separated (self.lhslist): |w{self.lhslist}|n\n"
        string += f"rhs, right-hand side of '=' (self.rhs): |w{self.rhs}|n\n"
        string += f"rhs, comma separated (self.rhslist): |w{self.rhslist}|n\n"
        string += "-" * 50
        self.msg(string)

    def func(self):
        """
        This is the hook function that actually does all the work. It is called
         by the cmdhandler right after self.parser() finishes, and so has access
         to all the variables defined therein.
        """
        self.get_command_info()


class MuxAccountCommand(MuxCommand):
    """
    This is an on-Account version of the MuxCommand. Since these commands sit
    on Accounts rather than on Characters/Objects, we need to check
    this in the parser.

    Account commands are available also when puppeting a Character, it's
    just that they are applied with a lower priority and are always
    available, also when disconnected from a character (i.e. "ooc").

    This class makes sure that caller is always an Account object, while
    creating a new property "character" that is set only if a
    character is actually attached to this Account and Session.
    """

    account_caller = True  # Using MuxAccountCommand explicitly defaults the caller to an account



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scripts", "0002_auto_20150118_1625")]

    operations = [
        migrations.CreateModel(
            name="ScriptBase", fields=[], options={"proxy": True}, bases=("scripts.scriptdb",)
        ),
        migrations.CreateModel(
            name="DefaultScript", fields=[], options={"proxy": True}, bases=("scripts.scriptbase",)
        ),
        migrations.CreateModel(
            name="DoNothing", fields=[], options={"proxy": True}, bases=("scripts.defaultscript",)
        ),
        migrations.CreateModel(
            name="CheckSessions",
            fields=[],
            options={"proxy": True},
            bases=("scripts.defaultscript",),
        ),
        migrations.CreateModel(
            name="Store", fields=[], options={"proxy": True}, bases=("scripts.defaultscript",)
        ),
        migrations.CreateModel(
            name="ValidateChannelHandler",
            fields=[],
            options={"proxy": True},
            bases=("scripts.defaultscript",),
        ),
        migrations.CreateModel(
            name="ValidateIdmapperCache",
            fields=[],
            options={"proxy": True},
            bases=("scripts.defaultscript",),
        ),
        migrations.CreateModel(
            name="ValidateScripts",
            fields=[],
            options={"proxy": True},
            bases=("scripts.defaultscript",),
        ),
    ]



"""
Default prototypes for building the XYZ-grid into actual game-rooms.

Add this to mygame/conf/settings/settings.py:

    PROTOTYPE_MODULES += ['evennia.contrib.grid.xyzgrid.prototypes']

The prototypes can then be used in mapping prototypes as

    {'prototype_parent': 'xyz_room', ...}

and/or

    {'prototype_parent': 'xyz_exit', ...}

"""
from django.conf import settings

try:
    room_override = settings.XYZROOM_PROTOTYPE_OVERRIDE
except AttributeError:
    room_override = {}

try:
    exit_override = settings.XYZEXIT_PROTOTYPE_OVERRIDE
except AttributeError:
    exit_override = {}

room_prototype = {
    "prototype_key": "xyz_room",
    "typeclass": "evennia.contrib.grid.xyzgrid.xyzroom.XYZRoom",
    "prototype_tags": ("xyzroom",),
    "key": "A room",
    "desc": "An empty room.",
}
room_prototype.update(room_override)

exit_prototype = {
    "prototype_key": "xyz_exit",
    "typeclass": "evennia.contrib.grid.xyzgrid.xyzroom.XYZExit",
    "prototype_tags": ("xyzexit",),
    "desc": "An exit.",
}
exit_prototype.update(exit_override)

# accessed by the prototype importer
PROTOTYPE_LIST = [room_prototype, exit_prototype]



"""
test_prototypes

This is meant to be used with Evennia unittest framework
to provide some named prototypes for use with various tests
(for example commands that accept named prototypes, not just
prototype dictionaries)
"""

GOBLIN = {"key": "Goblin", "typeclass": "DefaultCharacter", "desc": "A goblin."}

BALL = {"key": "Ball", "typeclass": "DefaultObject", "desc": "A ball."}



"""
Custom Evennia launcher command option for maintaining the grid in a separate process than the main
server (since this can be slow).

To use, add to the settings:
::

    EXTRA_LAUNCHER_COMMANDS.update({'xyzgrid': 'evennia.contrib.grid.xyzgrid.launchcmd.xyzcommand'})

You should now be able to do
::

    evennia xyzgrid <options>

Use `evennia xyzgrid help` for usage help.

"""

from os.path import join as pathjoin

from django.conf import settings

import evennia
from evennia.contrib.grid.xyzgrid.xyzgrid import get_xyzgrid
from evennia.utils import ansi

_HELP_SHORT = """
evennia xyzgrid help | list | init | add | spawn | initpath | delete [<options>]
 Manages the XYZ grid. Use 'xyzgrid help <option>' for documentation.
"""

_HELP_HELP = """
evennia xyzgrid <command> [<options>]
Manages the XYZ grid.

help <command>   - get help about each command:
    list            - show list
    init            - initialize grid (only one time)
    add             - add new maps to grid
    spawn           - spawn added maps into actual db-rooms/exits
    initpath        - (re)creates pathfinder matrices
    delete          - delete part or all of grid
"""

_HELP_LIST = """
list

    Lists the map grid structure and any loaded maps.

list <Z|mapname>

    Display the given XYmap in more detail. Also 'show' works. Use quotes around
    map-names with spaces.

Examples:

    evennia xyzgrid list
    evennia xyzgrid list mymap
    evennia xyzgrid list "the small cave"
"""

_HELP_INIT = """
init

    First start of the grid. This will create the XYZGrid global script. No maps are loaded yet!
    It's safe to run this command multiple times; the grid will only be initialized once.

Example:

    evennia xyzgrid init
"""


_HELP_ADD = """
add <path.to.xymap.module> [<path> <path>,...]

    Add path(s) to one or more modules containing XYMap definitions. The module will be parsed
    for

    - a XYMAP_DATA - a dict on this form:
        {"map": mapstring, "zcoord": mapname/zcoord, "legend": dict, "prototypes": dict}
        describing one single XYmap, or
    - a XYMAP_DATA_LIST - a list of multiple dicts on the XYMAP_DATA form. This allows for
        embedding multiple maps in the same module. See evennia/contrib/grid/xyzgrid/example.py
        for an example of how this looks.

    Note that adding a map does *not* spawn it. If maps are linked to one another, you should
    add all linked maps before running 'spawn', or you'll get errors when creating transitional
    exits between maps.

Examples:

    evennia xyzgrid add evennia.contrib.grid.xyzgrid.example
    evennia xyzgrid add world.mymap1 world.mymap2 world.mymap3
"""

_HELP_SPAWN = """
spawn

    spawns/updates the entire database grid based on the added maps. For a new grid, this will
    spawn all new rooms/exits (and may take a good while!). For updating, rooms may be
    removed/spawned if a map changed since the last spawn.

spawn "(X,Y,Z|mapname)"

    spawns/updates only a part of the grid. Remember the quotes around the coordinate (this
    is mostly because shells don't like them)! Use '*' as a wild card for XY coordinates.
    This should usually only be used if the full grid has already been built once - otherwise
    inter-map transitions may fail! Z is the name/z-coordinate of the map to spawn.

Examples:

    evennia xyzgrid spawn                  - spawn all
    evennia xyzgrid "(*, *, mymap1)"       - spawn everything of map/zcoord mymap1
    evennia xyzgrid "(12, 5, mymap1)"      - spawn only coordinate (12, 5) on map/zcoord mymap1
"""

_HELP_INITPATH = """
initpath

    Recreates the pathfinder matrices for the entire grid. These are used for all shortest-path
    calculations. The result will be cached to disk (in mygame/server/.cache/). If not run, each
    map will run this automatically first time it's used. Running this will always force to
    respawn the cache.

initpath Z|mapname

    recreate the pathfinder matrix for a specific map only. Z is the name/z-coordinate of the
    map. If the map name has spaces in it, use quotes.

Examples:

    evennia xyzgrid initpath
    evennia xyzgrid initpath mymap1
    evennia xyzgrid initpath "the small cave"
"""

_HELP_DELETE = """
delete

    WARNING: This will delete the entire xyz-grid (all maps), and *all* rooms/exits built to
    match it (they serve no purpose without the grid). You will be asked to confirm before
    continuing with this operation.

delete Z|mapname

    Remove a previously added XYmap with the name/z-coordinate Z. If the map was built, this
    will also wipe all its spawned rooms/exits. You will be asked to confirm before continuing
    with this operation. Use quotes if the Z/mapname contains spaces.

Examples:

    evennia xyzgrid delete
    evennia xyzgrid delete mymap1
    evennia xyzgrid delete "the small cave"
"""

_TOPICS_MAP = {
    "list": _HELP_LIST,
    "init": _HELP_INIT,
    "add": _HELP_ADD,
    "spawn": _HELP_SPAWN,
    "initpath": _HELP_INITPATH,
    "delete": _HELP_DELETE,
}

evennia._init()


def _option_help(*suboptions):
    """
    Show help <command> aid.

    """
    if not suboptions:
        topic = _HELP_HELP
    else:
        topic = _TOPICS_MAP.get(suboptions[0], _HELP_HELP)
    print(topic.strip())


def _option_list(*suboptions):
    """
    List/view grid.

    """

    xyzgrid = get_xyzgrid()

    # override grid's logger to echo directly to console
    def _log(msg):
        print(msg)

    xyzgrid.log = _log

    xymap_data = xyzgrid.grid
    if not xymap_data:
        if xyzgrid.db.map_data:
            print("Grid could not load due to errors.")
        else:
            print("The XYZgrid is currently empty. Use 'add' to add paths to your map data.")
        return

    if not suboptions:
        print("XYMaps stored in grid:")
        for zcoord, xymap in sorted(xymap_data.items(), key=lambda tup: tup[0]):
            print("\n" + str(repr(xymap)) + ":\n")
            print(ansi.parse_ansi(str(xymap)))
        return

    zcoord = " ".join(suboptions)
    xymap = xyzgrid.get_map(zcoord)
    if not xymap:
        print(f"No XYMap with Z='{zcoord}' was found on grid.")
    else:
        nrooms = xyzgrid.get_room(("*", "*", zcoord)).count()
        nnodes = len(xymap.node_index_map)
        print("\n" + str(repr(xymap)) + ":\n")
        checkwarning = True
        if not nrooms:
            print(f"{nrooms} / {nnodes} rooms are spawned.")
            checkwarning = False
        elif nrooms < nnodes:
            print(
                f"{nrooms} / {nnodes} rooms are spawned\n"
                "Note: Transitional nodes are *not* spawned (they just point \n"
                "to another map), so the 'missing room(s)' may just be from such nodes."
            )
        elif nrooms > nnodes:
            print(
                f"{nrooms} / {nnodes} rooms are spawned\n"
                "Note: Maybe some rooms were removed from map. Run 'spawn' to re-sync."
            )
        else:
            print(f"{nrooms} / {nnodes} rooms are spawned\n")

        if checkwarning:
            print(
                "Note: This check is not complete; it does not consider changed map "
                "topology\nlike relocated nodes/rooms and new/removed links/exits - this "
                "is calculated only during a spawn."
            )
        print("\nDisplayed map (as appearing in-game):\n\n" + ansi.parse_ansi(str(xymap)))
        print(
            "\nRaw map string (including axes and invisible nodes/links):\n" + str(xymap.mapstring)
        )
        print(f"\nCustom map options: {xymap.options}\n")
        legend = []
        for key, node_or_link in xymap.legend.items():
            legend.append(f"{key} - {node_or_link.__doc__.strip()}")
        print("Legend (all elements may not be present on map):\n " + "\n ".join(legend))


def _option_init(*suboptions):
    """
    Initialize a new grid. Will fail if a Grid already exists.

    """
    grid = get_xyzgrid()
    print(f"The grid is initalized as the Script '{grid.key}'({grid.dbref})")


def _option_add(*suboptions):
    """
    Add one or more map to the grid. Supports `add path,path,path,...`

    """
    grid = get_xyzgrid()

    # override grid's logger to echo directly to console
    def _log(msg):
        print(msg)

    grid.log = _log

    xymap_data_list = []
    for path in suboptions:
        maps = grid.maps_from_module(path)
        if not maps:
            print(f"No maps found with the path {path}.\nSeparate multiple paths with spaces. ")
            return
        mapnames = "\n ".join(f"'{m['zcoord']}'" for m in maps)
        print(f" XYMaps from {path}:\n {mapnames}")
        xymap_data_list.extend(maps)
    grid.add_maps(*xymap_data_list)
    try:
        grid.reload()
    except Exception as err:
        print(err)
    else:
        print(f"Added (or readded) {len(xymap_data_list)} XYMaps to grid.")


def _option_spawn(*suboptions):
    """
    spawn the grid or part of it.

    """
    grid = get_xyzgrid()

    # override grid's logger to echo directly to console
    def _log(msg):
        print(msg)

    grid.log = _log

    if suboptions:
        opts = "".join(suboptions).strip("()")
        # coordinate tuple
        try:
            x, y, z = (part.strip() for part in opts.split(","))
        except ValueError:
            print(
                "spawn coordinate must be given as (X, Y, Z) tuple, where '*' act "
                "wild cards and Z is the mapname/z-coord of the map to load."
            )
            return
    else:
        x, y, z = "*", "*", "*"

    if x == y == z == "*":
        inp = input(
            "This will (re)spawn the entire grid. If it was built before, it may spawn \n"
            "new rooms or delete rooms that no longer matches the grid.\nDo you want to "
            "continue? [Y]/N? "
        )
    else:
        inp = input(
            "This will spawn/delete objects in the database matching grid coordinates \n"
            f"({x},{y},{z}) (where '*' is a wildcard).\nDo you want to continue? [Y]/N? "
        )
    if inp.lower() in ("no", "n"):
        print("Aborted.")
        return

    print("Starting spawn ...")
    grid.spawn(xyz=(x, y, z))
    print(
        "... spawn complete!\nIt's recommended to reload the server to refresh caches if this "
        "modified an existing grid."
    )


def _option_initpath(*suboptions):
    """
    (Re)Initialize the pathfinding matrices for grid or part of it.

    """
    grid = get_xyzgrid()

    # override grid's logger to echo directly to console
    def _log(msg):
        print(msg)

    grid.log = _log

    xymaps = grid.all_maps()
    nmaps = len(xymaps)
    for inum, xymap in enumerate(xymaps):
        print(f"(Re)building pathfinding matrix for xymap Z={xymap.Z} ({inum+1}/{nmaps}) ...")
        xymap.calculate_path_matrix(force=True)

    cachepath = pathjoin(settings.GAME_DIR, "server", ".cache")
    print(f"... done. Data cached to {cachepath}.")


def _option_delete(*suboptions):
    """
    Delete the grid or parts of it. Allows mapname,mapname, ...

    """

    grid = get_xyzgrid()

    # override grid's logger to echo directly to console
    def _log(msg):
        print(msg)

    grid.log = _log

    if not suboptions:
        repl = input(
            "WARNING: This will delete the ENTIRE Grid and wipe all rooms/exits!"
            "\nObjects/Chars inside deleted rooms will be moved to their home locations."
            "\nThis can't be undone. Are you sure you want to continue? Y/[N]? "
        )
        if repl.lower() not in ("yes", "y"):
            print("Aborted.")
            return
        print("Deleting grid ...")
        grid.delete()
        print(
            "... done.\nPlease reload the server now; otherwise removed rooms may linger in cache."
        )
        return

    zcoords = (part.strip() for part in suboptions)
    err = False
    for zcoord in zcoords:
        if not grid.get_map(zcoord):
            print(f"Mapname/zcoord {zcoord} is not a part of the grid.")
            err = True
    if err:
        print("Valid mapnames/zcoords are\n:", "\n ".join(xymap.Z for xymap in grid.all_maps()))
        return
    repl = input(
        "This will delete map(s) {', '.join(zcoords)} and wipe all corresponding\n"
        "rooms/exits!"
        "\nObjects/Chars inside deleted rooms will be moved to their home locations."
        "\nThis can't be undone. Are you sure you want to continue? Y/[N]? "
    )
    if repl.lower() not in ("yes", "y"):
        print("Aborted.")
        return

    print("Deleting selected xymaps ...")
    grid.remove_map(*zcoords, remove_objects=True)
    print(
        "... done.\nPlease reload the server to refresh room caches."
        "\nAlso remember to remove any links from remaining maps pointing to deleted maps."
    )


def xyzcommand(*args):
    """
    Evennia launcher command. This is made available as `evennia xyzgrid` on the command line,
    once added to `settings.EXTRA_LAUNCHER_COMMANDS`.

    """
    if not args:
        print(_HELP_SHORT.strip())
        return

    option, *suboptions = args

    if option in ("help", "h"):
        _option_help(*suboptions)
    elif option in ("list", "show"):
        _option_list(*suboptions)
    elif option == "init":
        _option_init(*suboptions)
    elif option == "add":
        _option_add(*suboptions)
    elif option == "spawn":
        _option_spawn(*suboptions)
    elif option == "initpath":
        _option_initpath(*suboptions)
    elif option == "delete":
        _option_delete(*suboptions)
    else:
        print(f"Unknown option '{option}'. Use 'evennia xyzgrid help' for valid arguments.")



# -*- coding: utf-8 -*-
"""
EvMore - pager mechanism

This is a pager for displaying long texts and allows stepping up and down in
the text (the name comes from the traditional 'more' unix command).

To use, simply pass the text through the EvMore object:


```python

    from evennia.utils.evmore import EvMore

    text = some_long_text_output()
    EvMore(caller, text, always_page=False, session=None, justify_kwargs=None, **kwargs)
```

One can also use the convenience function `msg` from this module to avoid
having to set up the `EvMenu` object manually:

```python

    from evennia.utils import evmore

    text = some_long_text_output()
    evmore.msg(caller, text, always_page=False, session=None, justify_kwargs=None, **kwargs)
```

The `always_page` argument  decides if the pager is used also if the text is not long
enough to need to scroll, `session` is used to determine which session to relay
to and `justify_kwargs` are kwargs to pass to utils.utils.justify in order to
change the formatting of the text. The remaining `**kwargs` will be passed on to
the `caller.msg()` construct every time the page is updated.

----

"""
from django.conf import settings
from django.core.paginator import Paginator
from django.db.models.query import QuerySet
from django.utils.translation import gettext as _

import evennia
from evennia.commands import cmdhandler
from evennia.commands.cmdset import CmdSet
from evennia.commands.command import Command
from evennia.utils.ansi import ANSIString
from evennia.utils.utils import dedent, inherits_from, justify, make_iter

_CMD_NOMATCH = cmdhandler.CMD_NOMATCH
_CMD_NOINPUT = cmdhandler.CMD_NOINPUT

# we need to use NAWS for this
_SCREEN_WIDTH = settings.CLIENT_DEFAULT_WIDTH
_SCREEN_HEIGHT = settings.CLIENT_DEFAULT_HEIGHT

_EVTABLE = None

_LBR = ANSIString("\n")

# text

_DISPLAY = """{text}
|n(|wPage|n [{pageno}/{pagemax}] |wn|next|n || |wp|nrevious || |wt|nop || |we|nnd || |wq|nuit)"""


class CmdMore(Command):
    """
    Manipulate the text paging. Catch no-input with aliases.
    """

    key = _CMD_NOINPUT
    aliases = ["quit", "q", "abort", "a", "next", "n", "previous", "p", "top", "t", "end", "e"]
    auto_help = False

    def func(self):
        """
        Implement the command
        """
        more = self.caller.ndb._more
        if not more and inherits_from(self.caller, evennia.DefaultObject):
            more = self.caller.account.ndb._more
        if not more:
            self.caller.msg("Error in loading the pager. Contact an admin.")
            return

        cmd = self.cmdstring

        if cmd in ("abort", "a", "q"):
            more.page_quit()
        elif cmd in ("previous", "p"):
            more.page_back()
        elif cmd in ("top", "t", "look", "l"):
            more.page_top()
        elif cmd in ("end", "e"):
            more.page_end()
        else:
            # return or n, next
            more.page_next()


class CmdMoreExit(Command):
    """
    Any non-more command will exit the pager.

    """

    key = _CMD_NOMATCH

    def func(self):
        """
        Exit pager and re-fire the failed command.

        """
        more = self.caller.ndb._more
        more.page_quit()

        # re-fire the command (in new cmdset)
        self.caller.execute_cmd(self.raw_string)


class CmdSetMore(CmdSet):
    """
    Stores the more command
    """

    key = "more_commands"
    priority = 110
    mergetype = "Replace"

    def at_cmdset_creation(self):
        self.add(CmdMore())
        self.add(CmdMoreExit())


# resources for handling queryset inputs
def queryset_maxsize(qs):
    return qs.count()


class EvMore(object):
    """
    The main pager object
    """

    def __init__(
        self,
        caller,
        inp,
        always_page=False,
        session=None,
        justify=False,
        justify_kwargs=None,
        exit_on_lastpage=False,
        exit_cmd=None,
        page_formatter=str,
        **kwargs,
    ):
        """
        Initialization of the EvMore pager.

        Args:
            caller (Object or Account): Entity reading the text.
            inp (str, EvTable, Paginator or iterator): The text or data to put under paging.

                - If a string, paginage normally. If this text contains
                  one or more `\\\\f` format symbol, automatic pagination and justification
                  are force-disabled and page-breaks will only happen after each `\\\\f`.
                - If `EvTable`, the EvTable will be paginated with the same
                  setting on each page if it is too long. The table
                  decorations will be considered in the size of the page.
                - Otherwise `inp` is converted to an iterator, where each step is
                  expected to be a line in the final display. Each line
                  will be run through `iter_callable`.

            always_page (bool, optional): If `False`, the
                pager will only kick in if `inp` is too big
                to fit the screen.
            session (Session, optional): If given, this session will be used
                to determine the screen width and will receive all output.
            justify (bool, optional): If set, auto-justify long lines. This must be turned
                off for fixed-width or formatted output, like tables. It's force-disabled
                if `inp` is an EvTable.
            justify_kwargs (dict, optional): Keywords for the justify function. Used only
                if `justify` is True. If this is not set, default arguments will be used.
            exit_on_lastpage (bool, optional): If reaching the last page without the
                page being completely filled, exit pager immediately. If unset,
                another move forward is required to exit. If set, the pager
                exit message will not be shown.
            exit_cmd (str, optional): If given, this command-string will be executed on
                the caller when the more page exits. Note that this will be using whatever
                cmdset the user had *before* the evmore pager was activated (so none of
                the evmore commands will be available when this is run).
            kwargs (any, optional): These will be passed on to the `caller.msg` method. Notably,
                one can pass additional outputfuncs this way. There is one special kwarg:
                - `text_kwargs` - extra kwargs to pass with the text outputfunc, e.g.
                `text_kwargs={"type": "help"}` would result to each page being sent
                to `msg` as `text=(pagetxt, {"type": "help"})`.

        Examples:

            ```python
            super_long_text = " ... "
            EvMore(caller, super_long_text)
            ```
            Paginator
            ```python
            from django.core.paginator import Paginator
            query = ObjectDB.objects.all()
            pages = Paginator(query, 10)  # 10 objs per page
            EvMore(caller, pages)
            ```
            Every page an EvTable
            ```python
            from evennia import EvTable
            def _to_evtable(page):
                table = ... # convert page to a table
                return EvTable(*headers, table=table, ...)
            EvMore(caller, pages, page_formatter=_to_evtable)
            ```

        """
        self._caller = caller
        self._always_page = always_page

        if not session:
            # if not supplied, use the first session to
            # determine screen size
            sessions = caller.sessions.get()
            if not sessions:
                return
            session = sessions[0]
        self._session = session

        self._justify = justify
        self._justify_kwargs = justify_kwargs
        self.exit_on_lastpage = exit_on_lastpage
        self.exit_cmd = exit_cmd
        self._exit_msg = _("|xExited pager.|n")

        self._text_kwargs = kwargs.pop("text_kwargs", {})

        self._kwargs = kwargs

        self._data = None

        self._pages = []
        self._npos = 0

        self._npages = 1
        self._paginator = self.paginator_index
        self._page_formatter = str

        # set up individual pages for different sessions
        height = max(4, session.protocol_flags.get("SCREENHEIGHT", {0: _SCREEN_HEIGHT})[0] - 4)
        self.width = session.protocol_flags.get("SCREENWIDTH", {0: _SCREEN_WIDTH})[0]
        # always limit number of chars to 10 000 per page
        self.height = min(10000 // max(1, self.width), height)

        # does initial parsing of input
        self.init_pages(inp)

        # kick things into gear
        self.start()

    # EvMore functional methods

    def display(self, show_footer=True):
        """
        Pretty-print the page.
        """
        pos = 0
        text = "[no content]"
        if self._npages > 0:
            pos = self._npos
            text = self.page_formatter(self.paginator(pos))
        if show_footer:
            page = _DISPLAY.format(text=text, pageno=pos + 1, pagemax=self._npages)
        else:
            page = text
        # check to make sure our session is still valid
        sessions = self._caller.sessions.get()
        if not sessions:
            self.page_quit()
            return
        # this must be an 'is' check, not an == check
        if not any(ses for ses in sessions if self._session is ses):
            self._session = sessions[0]
        text_outputfunc = (page, (), self._text_kwargs)
        self._caller.msg(text=text_outputfunc, session=self._session, **self._kwargs)

    def page_top(self):
        """
        Display the top page
        """
        self._npos = 0
        self.display()

    def page_end(self):
        """
        Display the bottom page.
        """
        self._npos = self._npages - 1
        self.display()

    def page_next(self):
        """
        Scroll the text to the next page. Quit if already at the end
        of the page.
        """
        if self._npos >= self._npages - 1:
            # exit if we are already at the end
            self.page_quit()
        else:
            self._npos += 1
            if self.exit_on_lastpage and self._npos >= (self._npages - 1):
                self.display(show_footer=False)
                self.page_quit(quiet=True)
            else:
                self.display()

    def page_back(self):
        """
        Scroll the text back up, at the most to the top.
        """
        self._npos = max(0, self._npos - 1)
        self.display()

    def page_quit(self, quiet=False):
        """
        Quit the pager
        """
        del self._caller.ndb._more
        if not quiet:
            self._caller.msg(text=self._exit_msg, **self._kwargs)
        self._caller.cmdset.remove(CmdSetMore)
        if self.exit_cmd:
            self._caller.execute_cmd(self.exit_cmd, session=self._session)

    def start(self):
        """
        Starts the pagination
        """
        if self._npages <= 1 and not self._always_page:
            # no need for paging; just pass-through.
            self.display(show_footer=False)
        else:
            # go into paging mode
            # first pass on the msg kwargs
            self._caller.ndb._more = self
            self._caller.cmdset.add(CmdSetMore)

            # goto top of the text
            self.page_top()

    # default paginators - responsible for extracting a specific page number

    def paginator_index(self, pageno):
        """Paginate to specific, known index"""
        return self._data[pageno]

    def paginator_slice(self, pageno):
        """
        Paginate by slice. This is done with an eye on memory efficiency (usually for
        querysets); to avoid fetching all objects at the same time.

        """
        return self._data[pageno * self.height : pageno * self.height + self.height]

    def paginator_django(self, pageno):
        """
        Paginate using the django queryset Paginator API. Note that his is indexed from 1.
        """
        return self._data.page(pageno + 1)

    # default helpers to set up particular input types

    def init_evtable(self, table):
        """The input is an EvTable."""
        if table.height:
            # enforced height of each paged table, plus space for evmore extras
            self.height = table.height - 4

        # convert table to string
        text = str(table)
        self._justify = False
        self._justify_kwargs = None  # enforce
        self.init_str(text)

    def init_queryset(self, qs):
        """The input is a queryset"""
        nsize = qs.count()  # we assume each will be a line
        self._npages = nsize // self.height + (0 if nsize % self.height == 0 else 1)
        self._data = qs

    def init_django_paginator(self, pages):
        """
        The input is a django Paginator object.
        """
        self._npages = pages.num_pages
        self._data = pages

    def init_iterable(self, inp):
        """The input is something other than a string - convert to iterable of strings"""
        inp = make_iter(inp)
        nsize = len(inp)
        self._npages = nsize // self.height + (0 if nsize % self.height == 0 else 1)
        self._data = inp

    def init_f_str(self, text):
        """
        The input contains `\\f` markers. We use `\\f` to indicate the user wants to
        enforce their line breaks on their own. If so, we do no automatic
        line-breaking/justification at all.

        Args:
            text (str): The string to format with f-markers.

        """
        self._data = text.split("\f")
        self._npages = len(self._data)

    def init_str(self, text):
        """The input is a string"""

        if self._justify:
            # we must break very long lines into multiple ones. Note that this
            # will also remove spurious whitespace.
            justify_kwargs = self._justify_kwargs or {}
            width = self._justify_kwargs.get("width", self.width)
            justify_kwargs["width"] = width
            justify_kwargs["align"] = self._justify_kwargs.get("align", "l")
            justify_kwargs["indent"] = self._justify_kwargs.get("indent", 0)

            lines = []
            for line in text.split("\n"):
                if len(line) > width:
                    lines.extend(justify(line, **justify_kwargs).split("\n"))
                else:
                    lines.append(line)
        else:
            # no justification. Simple division by line
            lines = text.split("\n")

        # note: If joining on ANSIString here, we will parse out || escapes into | too early,
        # meaning the protocol will later parse into color; better to leave things and only parse
        # once.
        self._data = [
            "\n".join(lines[i : i + self.height]) for i in range(0, len(lines), self.height)
        ]
        self._npages = len(self._data)

    # Hooks for customizing input handling and formatting (override in a child class)

    def init_pages(self, inp):
        """
        Initialize the pagination. By default, will analyze input type to determine
        how pagination automatically.

        Args:
            inp (any): Incoming data to be paginated. By default, handles pagination of
                strings, querysets, django.Paginator, EvTables and any iterables with strings.

        Notes:
            If overridden, this method must perform the following  actions:

            - read and re-store `self._data` (the incoming data set) if needed for pagination to
              work.
            - set `self._npages` to the total number of pages. Default is 1.
            - set `self._paginator` to a callable that will take a page number 1...N and return
              the data to display on that page (not any decorations or next/prev buttons). If only
              wanting to change the paginator, override `self.paginator` instead.
            - set `self._page_formatter` to a callable that will receive the page from
              `self._paginator` and format it with one element per line. Default is `str`. Or
              override `self.page_formatter` directly instead.

            By default, helper methods are called that perform these actions
            depending on supported inputs.

        """
        if inherits_from(inp, "evennia.utils.evtable.EvTable"):
            # an EvTable
            self.init_evtable(inp)
            self._paginator = self.paginator_index
        elif isinstance(inp, QuerySet):
            # a queryset
            self.init_queryset(inp)
            self._paginator = self.paginator_slice
        elif isinstance(inp, Paginator):
            self.init_django_paginator(inp)
            self._paginator = self.paginator_django
        elif not isinstance(inp, str):
            # anything else not a str
            self.init_iterable(inp)
            self._paginator = self.paginator_slice
        elif "\f" in inp:
            # string with \f line-break markers in it
            self.init_f_str(inp)
            self._paginator = self.paginator_index
        else:
            # a string
            self.init_str(inp)
            self._paginator = self.paginator_index

    def paginator(self, pageno):
        """
        Paginator. The data operated upon is in `self._data`.

        Args:
            pageno (int): The page number to view, from 0...N-1
        Returns:
            str: The page to display (without any decorations, those are added
                by EvMore).

        """
        return self._paginator(pageno)

    def page_formatter(self, page):
        """
        Page formatter. Every page passes through this method. Override
        it to customize behavior per-page. A common use is to generate a new
        EvTable for every page (this is more efficient than to generate one huge
        EvTable across many pages and feed it into EvMore all at once).

        Args:
            page (any): A piece of data representing one page to display. This must

        Returns:
            str: A ready-formatted page to display. Extra footer with help about
                switching to the next/prev page will be added automatically

        """
        return self._page_formatter(page)


# helper function


def msg(
    caller,
    text="",
    always_page=False,
    session=None,
    justify=False,
    justify_kwargs=None,
    exit_on_lastpage=True,
    **kwargs,
):
    """
    EvMore-supported version of msg, mimicking the normal msg method.

    """
    EvMore(
        caller,
        text,
        always_page=always_page,
        session=session,
        justify=justify,
        justify_kwargs=justify_kwargs,
        exit_on_lastpage=exit_on_lastpage,
        **kwargs,
    )


msg.__doc__ += dedent(EvMore.__init__.__doc__)



"""
Custom gametime contrib - Griatch, vlgeoff 2017

"""

from .custom_gametime import GametimeScript  # noqa
from .custom_gametime import custom_gametime  # noqa
from .custom_gametime import gametime_to_realtime  # noqa
from .custom_gametime import real_seconds_until  # noqa
from .custom_gametime import realtime_to_gametime  # noqa
from .custom_gametime import schedule  # noqa
from .custom_gametime import UNITS, time_to_tuple  # noqa



"""
Crafting - Griatch 2020

"""

from .crafting import CraftingRecipe  # noqa
from .crafting import CraftingRecipeBase  # noqa
from .crafting import craft  # noqa
from .crafting import CraftingError, CraftingValidationError  # noqa



# -*- coding: utf-8 -*-


import django.core.validators
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("auth", "0001_initial"), ("typeclasses", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="AccountDB",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="last login"
                    ),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        help_text="Required. 30 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        unique=True,
                        max_length=30,
                        verbose_name="username",
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[\\w.@+-]+$", "Enter a valid username.", "invalid"
                            )
                        ],
                    ),
                ),
                (
                    "first_name",
                    models.CharField(max_length=30, verbose_name="first name", blank=True),
                ),
                (
                    "last_name",
                    models.CharField(max_length=30, verbose_name="last name", blank=True),
                ),
                (
                    "email",
                    models.EmailField(max_length=75, verbose_name="email address", blank=True),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                ("db_key", models.CharField(max_length=255, verbose_name="key", db_index=True)),
                (
                    "db_typeclass_path",
                    models.CharField(
                        help_text="this defines what 'type' of entity this is. This variable holds a Python path to a module with a valid Evennia Typeclass.",
                        max_length=255,
                        null=True,
                        verbose_name="typeclass",
                    ),
                ),
                (
                    "db_date_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="creation date"),
                ),
                (
                    "db_lock_storage",
                    models.TextField(
                        help_text="locks limit access to an entity. A lock is defined as a 'lock string' on the form 'type:lockfunctions', defining what functionality is locked and how to determine access. Not defining a lock means no access is granted.",
                        verbose_name="locks",
                        blank=True,
                    ),
                ),
                (
                    "db_is_connected",
                    models.BooleanField(
                        default=False,
                        help_text="If account is connected to game or not",
                        verbose_name="is_connected",
                    ),
                ),
                (
                    "db_cmdset_storage",
                    models.CharField(
                        help_text="optional python path to a cmdset class. If creating a Character, this will default to settings.CMDSET_CHARACTER.",
                        max_length=255,
                        null=True,
                        verbose_name="cmdset",
                    ),
                ),
                (
                    "db_is_bot",
                    models.BooleanField(
                        default=False,
                        help_text="Used to identify irc/rss bots",
                        verbose_name="is_bot",
                    ),
                ),
                (
                    "db_attributes",
                    models.ManyToManyField(
                        help_text="attributes on this object. An attribute can hold any pickle-able python object (see docs for special cases).",
                        to="typeclasses.Attribute",
                        null=True,
                    ),
                ),
                (
                    "db_tags",
                    models.ManyToManyField(
                        help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                        to="typeclasses.Tag",
                        null=True,
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Group",
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of his/her group.",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        related_query_name="user",
                        related_name="user_set",
                        to="auth.Permission",
                        blank=True,
                        help_text="Specific permissions for this user.",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={"verbose_name": "Account", "verbose_name_plural": "Accounts"},
            bases=(models.Model,),
        )
    ]



"""
Start menu

This is started from the `evscaperoom` command.

Here player user can set their own description as well as select to create a
new room (to start from scratch) or join an existing room (with other players).

"""
from evennia import EvMenu
from evennia.utils import create, justify, list_to_string, logger
from evennia.utils.evmenu import list_node

from .room import EvscapeRoom
from .utils import create_fantasy_word

# ------------------------------------------------------------
# Main menu
# ------------------------------------------------------------

_START_TEXT = """
|mEv|rScape|mRoom|n

|x- an escape-room experience using Evennia|n

You are |c{name}|n - {desc}|n.

Make a selection below.
"""

_CREATE_ROOM_TEXT = """
This will create a |ynew, empty room|n to challenge you.

Other players can be thrown in there at any time.

Remember that if you give up and are the last person to leave, that particular
room will be gone!

|yDo you want to create (and automatically join) a new room?|n")
"""

_JOIN_EXISTING_ROOM_TEXT = """
This will have you join an existing room ({roomname}).

This is {percent}% complete and has {nplayers} player(s) in it already:

 {players}

|yDo you want to join this room?|n
"""


def _move_to_room(caller, raw_string, **kwargs):
    """
    Helper to move a user to a room

    """
    room = kwargs["room"]
    room.msg_char(caller, f"Entering room |c'{room.name}'|n ...")
    room.msg_room(caller, f"~You |c~were just tricked in here too!|n")
    # we do a manual move since we don't want all hooks to fire.
    old_location = caller.location
    caller.location = room
    room.at_object_receive(caller, old_location)
    return "node_quit", {"quiet": True}


def _create_new_room(caller, raw_string, **kwargs):
    # create a random name, retrying until we find
    # a unique one
    key = create_fantasy_word(length=5, capitalize=True)
    while EvscapeRoom.objects.filter(db_key=key):
        key = create_fantasy_word(length=5, capitalize=True)
    room = create.create_object(EvscapeRoom, key=key)
    # we must do this once manually for the new room
    room.statehandler.init_state()
    _move_to_room(caller, "", room=room)

    nrooms = EvscapeRoom.objects.all().count()
    logger.log_info(
        f"Evscaperoom: {caller.key} created room '{key}' (#{room.id}). Now {nrooms} room(s) active."
    )

    room.log(f"JOIN: {caller.key} created and joined room")
    return "node_quit", {"quiet": True}


def _get_all_rooms(caller):
    """
    Get a list of all available rooms and store the mapping
    between option and room so we get to it later.

    """
    room_option_descs = []
    room_map = {}
    for room in EvscapeRoom.objects.all():
        if not room.pk or room.db.deleting:
            continue
        stats = room.db.stats or {"progress": 0}
        progress = int(stats["progress"])
        nplayers = len(room.get_all_characters())
        desc = (
            f"Join room |c'{room.get_display_name(caller)}'|n "
            f"(complete: {progress}%, players: {nplayers})"
        )
        room_map[desc] = room
        room_option_descs.append(desc)
    caller.ndb._menutree.room_map = room_map
    return room_option_descs


def _select_room(caller, menuchoice, **kwargs):
    """
    Get a room from the selection using the mapping we created earlier.
    """
    room = caller.ndb._menutree.room_map[menuchoice]
    return "node_join_room", {"room": room}


@list_node(_get_all_rooms, _select_room)
def node_start(caller, raw_string, **kwargs):
    text = _START_TEXT.strip()
    text = text.format(name=caller.key, desc=caller.db.desc)

    # build a list of available rooms
    options = (
        {
            "key": (
                "|y[s]et your description|n",
                "set your description",
                "set",
                "desc",
                "description",
                "s",
            ),
            "goto": "node_set_desc",
        },
        {
            "key": ("|y[c]reate/join a new room|n", "create a new room", "create", "c"),
            "goto": "node_create_room",
        },
        {"key": ("|r[q]uit the challenge", "quit", "q"), "goto": "node_quit"},
    )

    return text, options


def node_set_desc(caller, raw_string, **kwargs):
    current_desc = kwargs.get("desc", caller.db.desc)

    text = (
        "Your current description is\n\n " f'   "{current_desc}"' "\n\nEnter your new description!"
    )

    def _temp_description(caller, raw_string, **kwargs):
        desc = raw_string.strip()
        if 5 < len(desc) < 40:
            return None, {"desc": raw_string.strip()}
        else:
            caller.msg("|rYour description must be 5-40 characters long.|n")
            return None

    def _set_description(caller, raw_string, **kwargs):
        caller.db.desc = kwargs.get("desc")
        caller.msg("You set your description!")
        return "node_start"

    options = (
        {"key": "_default", "goto": _temp_description},
        {"key": ("|g[a]ccept", "a"), "goto": (_set_description, {"desc": current_desc})},
        {"key": ("|r[c]ancel", "c"), "goto": "node_start"},
    )
    return text, options


def node_create_room(caller, raw_string, **kwargs):
    text = _CREATE_ROOM_TEXT

    options = (
        {"key": ("|g[c]reate new room and start game|n", "c"), "goto": _create_new_room},
        {"key": ("|r[a]bort and go back|n", "a"), "goto": "node_start"},
    )

    return text, options


def node_join_room(caller, raw_string, **kwargs):
    room = kwargs["room"]
    stats = room.db.stats or {"progress": 0}

    players = [char.key for char in room.get_all_characters()]
    text = _JOIN_EXISTING_ROOM_TEXT.format(
        roomname=room.get_display_name(caller),
        percent=int(stats["progress"]),
        nplayers=len(players),
        players=list_to_string(players),
    )

    options = (
        {"key": ("|g[a]ccept|n (default)", "a"), "goto": (_move_to_room, kwargs)},
        {"key": ("|r[c]ancel|n", "c"), "goto": "node_start"},
        {"key": "_default", "goto": (_move_to_room, kwargs)},
    )

    return text, options


def node_quit(caller, raw_string, **kwargs):
    quiet = kwargs.get("quiet")
    text = ""
    if not quiet:
        text = "Goodbye for now!\n"
        # we check an Attribute on the caller to see if we should
        # leave the game entirely when leaving
        if caller.db.evscaperoom_standalone:
            from evennia import default_cmds
            from evennia.commands import cmdhandler

            cmdhandler.cmdhandler(
                caller.ndb._menutree._session, "", cmdobj=default_cmds.CmdQuit(), cmdobj_key="@quit"
            )

    return text, None  # empty options exit the menu


class EvscaperoomMenu(EvMenu):
    """
    Custom menu with a different formatting of options.

    """

    node_border_char = "~"

    def nodetext_formatter(self, text):
        return justify(text.strip("\n").rstrip(), align="c", indent=1)

    def options_formatter(self, optionlist):
        main_options = []
        room_choices = []
        for key, desc in optionlist:
            if key.isdigit():
                room_choices.append((key, desc))
            else:
                main_options.append(key)
        main_options = " | ".join(main_options)
        room_choices = super().options_formatter(room_choices)
        return "{}{}{}".format(main_options, "\n\n" if room_choices else "", room_choices)


# access function
def run_evscaperoom_menu(caller):
    """
    Run room selection menu

    """
    menutree = {
        "node_start": node_start,
        "node_quit": node_quit,
        "node_set_desc": node_set_desc,
        "node_create_room": node_create_room,
        "node_join_room": node_join_room,
    }

    EvscaperoomMenu(caller, menutree, startnode="node_start", cmd_on_exit=None, auto_quit=True)


# ------------------------------------------------------------
# In-game Options menu
# ------------------------------------------------------------


def _set_thing_style(caller, raw_string, **kwargs):
    room = caller.location
    options = caller.attributes.get("options", category=room.tagcategory, default={})
    options["things_style"] = kwargs.get("value", 2)
    caller.attributes.add("options", options, category=room.tagcategory)
    return None, kwargs  # rerun node


def _toggle_screen_reader(caller, raw_string, **kwargs):
    session = kwargs["session"]
    # flip old setting
    session.protocol_flags["SCREENREADER"] = not session.protocol_flags.get("SCREENREADER", False)
    # sync setting with portal
    session.sessionhandler.session_portal_sync(session)
    return None, kwargs  # rerun node


def node_options(caller, raw_string, **kwargs):
    text = "|cOption menu|n\n('|wq|nuit' to return)"
    room = caller.location

    options = caller.attributes.get("options", category=room.tagcategory, default={})
    things_style = options.get("things_style", 2)

    session = kwargs["session"]  # we give this as startnode_input when starting menu
    screenreader = session.protocol_flags.get("SCREENREADER", False)

    options = (
        {
            "desc": "{}No item markings (hard mode)".format(
                "|g(*)|n " if things_style == 0 else "( ) "
            ),
            "goto": (_set_thing_style, {"value": 0, "session": session}),
        },
        {
            "desc": "{}Items marked as |yitem|n (with color)".format(
                "|g(*)|n " if things_style == 1 else "( ) "
            ),
            "goto": (_set_thing_style, {"value": 1, "session": session}),
        },
        {
            "desc": "{}Items are marked as |y[item]|n (screenreader friendly)".format(
                "|g(*)|n " if things_style == 2 else "( ) "
            ),
            "goto": (_set_thing_style, {"value": 2, "session": session}),
        },
        {
            "desc": "{}Screenreader mode".format("(*) " if screenreader else "( ) "),
            "goto": (_toggle_screen_reader, kwargs),
        },
    )
    return text, options


class OptionsMenu(EvMenu):
    """
    Custom display of Option menu
    """

    def node_formatter(self, nodetext, optionstext):
        return f"{nodetext}\n\n{optionstext}"


# access function
def run_option_menu(caller, session):
    """
    Run option menu in-game
    """
    menutree = {"node_start": node_options}

    OptionsMenu(
        caller,
        menutree,
        startnode="node_start",
        cmd_on_exit="look",
        auto_quit=True,
        startnode_input=("", {"session": session}),
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2019-01-30 20:56

from base64 import b64decode, b64encode
from copy import deepcopy
from pickle import dumps, loads

from django.db import migrations

from evennia.utils.utils import to_bytes


def forwards(apps, schema_editor):
    Attribute = apps.get_model("typeclasses", "Attribute")
    for attr in Attribute.objects.all():
        # we need to re-assign the Attribute it's own value to make sure pickle switches from v2 to v4,
        # otherwise we will not be able to search db by-value.
        attr.db_value = attr.db_value
        attr.save(update_fields=["db_value"])


class Migration(migrations.Migration):

    dependencies = [("typeclasses", "0011_auto_20190128_1820")]

    operations = [migrations.RunPython(forwards, migrations.RunPython.noop)]



"""
FilterSets allow clients to specify querystrings that will determine the data
that is retrieved in GET requests. By default, Django Rest Framework uses the
'django-filter' package as its backend. Django-filter also has a section in its
documentation specifically regarding DRF integration.

https://django-filter.readthedocs.io/en/latest/guide/rest_framework.html

"""
from typing import Union

from django.db.models import Q
from django_filters.filters import EMPTY_VALUES, CharFilter
from django_filters.rest_framework.filterset import FilterSet

from evennia.accounts.models import AccountDB
from evennia.objects.models import ObjectDB
from evennia.scripts.models import ScriptDB


def get_tag_query(tag_type: Union[str, None], key: str) -> Q:
    """
    Returns a Q object for searching by tag names for typeclasses
    Args:
        tag_type(str or None): The type of tag (None, 'alias', etc)
        key (str): The name of the tag

    Returns:
        A Q object that for searching by this tag type and name

    """
    return Q(db_tags__db_tagtype=tag_type) & Q(db_tags__db_key__iexact=key)


class TagTypeFilter(CharFilter):
    """
    This class lets you create different filters for tags of a specified db_tagtype.

    """

    tag_type = None

    def filter(self, qs, value):
        # if no value is specified, we don't use the filter
        if value in EMPTY_VALUES:
            return qs
        # if they enter a value, we filter objects by having a tag of this type with the given name
        return qs.filter(get_tag_query(self.tag_type, value)).distinct()


class AliasFilter(TagTypeFilter):
    """A filter for objects by their aliases (tags with a tagtype of 'alias'"""

    tag_type = "alias"


class PermissionFilter(TagTypeFilter):
    """A filter for objects by their permissions (tags with a tagtype of 'permission'"""

    tag_type = "permission"


SHARED_FIELDS = ["db_key", "db_typeclass_path", "db_tags__db_key", "db_tags__db_category"]


class BaseTypeclassFilterSet(FilterSet):
    """
    A parent class with filters for aliases and permissions

    """

    name = CharFilter(lookup_expr="iexact", method="filter_name", field_name="db_key")
    alias = AliasFilter(lookup_expr="iexact")
    permission = PermissionFilter(lookup_expr="iexact")

    @staticmethod
    def filter_name(queryset, name, value):
        """
        Filters a queryset by aliases or the key of the typeclass
        Args:
            queryset: The queryset being filtered
            name: The name of the field
            value: The value passed in from GET params

        Returns:
            The filtered queryset
        """
        query = Q(**{f"{name}__iexact": value})
        query |= get_tag_query("alias", value)
        return queryset.filter(query).distinct()


class ObjectDBFilterSet(BaseTypeclassFilterSet):
    """
    This adds filters for ObjectDB instances - characters, rooms, exits, etc

    """

    class Meta:
        model = ObjectDB
        fields = SHARED_FIELDS + [
            "db_location__db_key",
            "db_home__db_key",
            "db_location__id",
            "db_home__id",
        ]


class AccountDBFilterSet(BaseTypeclassFilterSet):
    """This adds filters for Account objects"""

    name = CharFilter(lookup_expr="iexact", method="filter_name", field_name="username")

    class Meta:
        model = AccountDB
        fields = [
            fi
            for fi in (SHARED_FIELDS + ["username", "db_is_connected", "db_is_bot"])
            if fi != "db_key"
        ]


class ScriptDBFilterSet(BaseTypeclassFilterSet):
    """This adds filters for Script objects"""

    class Meta:
        model = ScriptDB
        fields = SHARED_FIELDS + [
            "db_desc",
            "db_obj__db_key",
            "db_obj__id",
            "db_account__id",
            "db_account__username",
            "db_is_active",
            "db_persistent",
            "db_interval",
        ]


class HelpFilterSet(FilterSet):

    """
    Filter for help entries

    """

    name = CharFilter(lookup_expr="iexact", method="filter_name", field_name="db_key")
    category = CharFilter(lookup_expr="iexact", method="filter_name", field_name="db_category")
    alias = AliasFilter(lookup_expr="iexact")



"""
First room state

This simple example sets up an empty one-state room with a door and a key.
After unlocking, opening the door and leaving the room, the player is
teleported back to the evscaperoom menu.


"""

from evennia.contrib.full_systems.evscaperoom import objects
from evennia.contrib.full_systems.evscaperoom.state import BaseState

GREETING = """
This is the situation, {name}:

You are locked in this room ... get out! Simple as that!

"""


ROOM_DESC = """
This is a featureless room. On one wall is a *door. On the other wall is a
*button marked "GET HELP".

There is a *key lying on the floor.

"""

# Example object

DOOR_DESC = """
This is a simple example door leading out of the room.

"""


class Door(objects.Openable):
    """
    The door leads out of the room.

    """

    start_open = False

    def at_object_creation(self):
        super().at_object_creation()
        self.set_flag("door")  # target for key

    def at_open(self, caller):
        # only works if the door was unlocked
        self.msg_room(caller, f"~You ~open *{self.key}")

    def at_focus_leave(self, caller, **kwargs):
        if self.check_flag("open"):
            self.msg_room(caller, "~You ~leave the room!")
            self.msg_char(caller, "Congrats!")
            # exit evscaperoom
            self.room.character_exit(caller)
        else:
            self.msg_char(caller, "The door is closed. You cannot leave!")


# key

KEY_DESC = """
A simple room key. A paper label is attached to it.
"""

KEY_READ = """
A paper label is attached to the key. It reads.

    |rOPEN THE DOOR WITH ME|n

A little on the nose, but this is an example room after all ...
"""

KEY_APPLY = """
~You insert the *key into the *door, turns it and ... the door unlocks!

"""


class Key(objects.Insertable, objects.Readable):
    "A key for opening the door"

    # where this key applies (must be flagged as such)
    target_flag = "door"

    def at_apply(self, caller, action, obj):
        obj.set_flag("unlocked")  # unlocks the door
        self.msg_room(caller, KEY_APPLY.strip())

    def at_read(self, caller, *args, **kwargs):
        self.msg_char(caller, KEY_READ.strip())

    def get_cmd_signatures(self):
        return [], "You can *read the label or *insert the key into something."


# help button

BUTTON_DESC = """
On the wall is a button marked

    PRESS ME FOR HELP

"""


class HelpButton(objects.EvscaperoomObject):
    def at_focus_push(self, caller, **kwargs):
        "this adds the 'push' action to the button"

        hint = self.room.state.get_hint()
        if hint is None:
            self.msg_char(caller, "There are no more hints to be had.")
        else:
            self.msg_room(
                caller, f"{caller.key} pushes *button and gets the " f'hint:\n "{hint.strip()}"|n'
            )


# state

STATE_HINT_LVL1 = """
The door is locked. What is usually used for unlocking doors?
"""

STATE_HINT_LVL2 = """
This is just an example. Do what comes naturally. Examine what's on the floor.
"""

STATE_HINT_LVL3 = """
Insert the *key in the *door. Then open the door and leave! Yeah, it's really
that simple.

"""


class State(BaseState):
    """
    This class (with exactly this name) must exist in every state module.

    """

    # this makes these hints available to the .get_hint method.
    hints = [STATE_HINT_LVL1, STATE_HINT_LVL2, STATE_HINT_LVL3]

    def character_enters(self, char):
        "Called when char enters room at this state"
        self.cinematic(GREETING.format(name=char.key))

    def init(self):
        "Initialize state"

        # describe the room
        self.room.db.desc = ROOM_DESC

        # create the room objects
        door = self.create_object(Door, key="door to the cabin", aliases=["door"])
        door.db.desc = DOOR_DESC.strip()

        key = self.create_object(Key, key="key", aliases=["room key"])
        key.db.desc = KEY_DESC.strip()

        button = self.create_object(HelpButton, key="button", aliases=["help button"])
        button.db.desc = BUTTON_DESC.strip()

    def clean(self):
        "Cleanup operations on the state, when it's over"
        super().clean()
